import rio

# <additional-imports>
from .. import data_models, persistence

# </additional-imports>


# <component>
class UserSignUpForm(rio.Component):
    """
    Provides interface for users to sign up for a new account.

    It includes fields for username and password, handles user creation, and
    displays error messages if the sign-up process fails.
    """

    # This will be set to `True` when the sign-up popup is open
    popup_open: bool

    # These fields will be bound to the input fields in the form. This allows us
    # to easily access the values entered by the user.
    username_sign_up: str = ""
    password_sign_up: str = ""
    password_sign_up_repeat: str = ""

    # This field will be used to display an error message if the sign-up process
    # fails.
    error_message: str = ""

    # These fields will be updated to reflect the validity of the username and
    # passwords entered by the user. If they are invalid, we will display an
    # error message to the user.
    username_valid: bool = True
    passwords_valid: bool = True

    async def on_sign_up(self) -> None:
        """
        Handles the sign-up process when the user submits the sign-up form.

        It will check if the user already exists and if the passwords match. If
        the user does not exist and the passwords match, a new user will be
        created and stored in the database.
        """
        # Get the persistence instance. It was attached to the session earlier,
        # so we can easily access it from anywhere.
        pers = self.session[persistence.Persistence]

        # Make sure all fields are populated
        if (
            not self.username_sign_up
            or not self.password_sign_up
            or not self.password_sign_up_repeat
        ):
            self.error_message = "Please fill in all fields"
            self.passwords_valid = False
            self.username_valid = False
            return

        # Check if the passwords match
        if self.password_sign_up != self.password_sign_up_repeat:
            self.error_message = "Passwords do not match"
            self.passwords_valid = False
            self.username_valid = True
            return

        # Check if this username is available
        try:
            await pers.get_user_by_username(username=self.username_sign_up)
        except KeyError:
            pass
        else:
            self.error_message = "This username is already taken"
            self.username_valid = False
            self.passwords_valid = True
            return

        # Create a new user
        user_info = data_models.AppUser.new_with_defaults(
            username=self.username_sign_up,
            password=self.password_sign_up,
        )

        # Store the user in the database
        await pers.create_user(user_info)

        # Registration is complete - close the popup
        self.popup_open = False

        # Log the user in, so they can start using the app straight away. To do
        # this, first create a session.
        user_session = await pers.create_session(
            user_id=user_info.id,
        )

        # Attach the session and userinfo. This indicates to any other
        # component in the app that somebody is logged in, and who that is.
        self.session.attach(user_session)
        self.session.attach(user_info)

        # Permanently store the session token with the connected client.
        # This way they can be recognized again should they reconnect later.
        settings = self.session[data_models.UserSettings]
        settings.auth_token = user_session.id
        self.session.attach(settings)

        # The user is logged in - no reason to stay here
        self.session.navigate_to("/app/home")

    def on_cancel(self) -> None:
        """
        Closes the sign-up popup when the user clicks the cancel button.

        It also resets all fields to their default values.
        """
        # Set all fields to default values
        self.username_valid: bool = True
        self.passwords_valid: bool = True
        self.username_sign_up: str = ""
        self.password_sign_up: str = ""
        self.password_sign_up_repeat: str = ""
        self.error_message: str = ""

        # Close pop up
        self.popup_open = False

    def build(self) -> rio.Component:
        # Determine the layout based on the window width
        desktop_layout = self.session.window_width > 30

        return rio.Card(
            rio.Column(
                # Heading
                rio.Text("Create a new account", style="heading1"),
                # Display an error, if any
                rio.Banner(
                    text=self.error_message,
                    style="danger",
                    margin_top=1,
                ),
                # Form fields
                rio.TextInput(
                    text=self.bind().username_sign_up,
                    label="Username*",
                    is_valid=self.username_valid,
                ),
                rio.TextInput(
                    text=self.bind().password_sign_up,
                    label="Password*",
                    is_valid=self.passwords_valid,
                    is_secret=True,
                ),
                rio.TextInput(
                    text=self.bind().password_sign_up_repeat,
                    label="Repeat Password*",
                    is_valid=self.passwords_valid,
                    is_secret=True,
                ),
                # And finally, some buttons to confirm or cancel the sign-up
                # process
                rio.Row(
                    rio.Button(
                        "Sign up",
                        on_press=self.on_sign_up,
                    ),
                    rio.Button(
                        "Cancel",
                        on_press=self.on_cancel,
                        style="minor",
                    ),
                    spacing=2,
                    margin_top=1,
                ),
                spacing=1,
                margin=2 if desktop_layout else 1,
            ),
            margin_x=0.5,
            align_x=0.5 if desktop_layout else None,
            min_width=24 if desktop_layout else 0,
            align_y=0.5,
        )


# </component>



# <additional-imports>
import rio

from .. import theme

# </additional-imports>


# <component>
class CustomSwitcherBar(rio.Component):
    """
    MySwitcher Component.

    This component provides a toggle switch allowing users to select between
    "Monthly" and "Yearly" durations. It is styled to blend seamlessly with the
    website's theme, featuring rounded corners and a transparent background with
    a subtle border.

    ## Attributes:

    `duartion`: The selected duration, either "Monthly" or "Yearly".
    """

    duartion: str = "Monthly"

    def build(self) -> rio.Component:
        return rio.Rectangle(
            # The switcher bar allowing users to select between "Monthly" and
            # "Yearly"
            content=rio.SwitcherBar(
                values=["Monthly", "Yearly"],
                # The selected value is bound to the "duration" attribute,
                # therefore, no eventhandler is needed
                selected_value=self.bind().duartion,
                color=rio.Color.from_hex("#ffffff"),
                align_x=0.5,  # Horizontally center the switcher within the rectangle
                min_width=10,  # Fixed width of the switcher bar
                margin=0.3,  # Margin around the switcher bar for spacing
            ),
            fill=rio.Color.TRANSPARENT,
            # Add a subtle border to the rectangle for visual separation
            stroke_width=0.1,
            stroke_color=theme.NEUTRAL_COLOR_BRIGHTER,
            # Fully rounded corners to create a pill-shaped rectangle.
            corner_radius=99999,
            align_x=0.5,  # Horizontally center the rectangle within its parent
        )


# </component>



from __future__ import annotations

import copy
import dataclasses
import typing as t

import imy.docstrings
import typing_extensions as te
import uniserde

from . import inspection, serialization, session
from .dataclass import RioDataclassMeta, all_class_fields

__all__ = [
    "UserSettings",
]


@imy.docstrings.mark_constructor_as_private  # Don't document the constructor
class UserSettings(metaclass=RioDataclassMeta):
    """
    Base class for persistent user settings.

    When creating an app or website you'll often want to store some values so
    that you can access them the next time the user visits your app. A typical
    example are configuration values set by the user - you wouldn't want to ask
    for these every time.

    Rio makes it easy to store and retrieve such values. Create a class that
    inherits from `UserSettings`, and attach it to the `Session`. That's it! Rio
    will automatically store and retrieve the values for you.

    ```python
    # Create a dataclass that inherits from rio.UserSettings. This indicates to
    # Rio that these are settings and should be persisted.
    class MySettings(rio.UserSettings):
        language: str = "en"

    # Attach the settings to the app. This way the settings will be available in
    # all sessions. They will be loaded automatically from the user whenever
    # they connect or start the app.
    app = rio.App(
        ...,
        default_attachments=[
            MySettings(),
        ],
    )
    ```

    You can modify the settings from anywhere in your app. Rio will detect
    changes and persist them automatically:

    ```python
    # ... somewhere in your code
    settings = self.session[MySettings]

    # Read any values you need to
    print(settings.language)  # "en"

    # Assignments will be automatically detected and saved
    settings.language = "de"
    ```

    Warning: Since settings are stored on the user's device, special
    considerations apply. Some countries have strict privacy laws regulating
    what you can store with/without the user's consent. Make sure you are
    familiar with the legal situation before going wild and storing everything
    you can think of.

    Warning: Since settings are stored on the user's device, you should never
    trust them to be valid. A malicious actor could modify them to intentionally
    trigger bugs in your app. Always validate the values before using them.

    ## Attributes

    `section_name`: If provided, the settings file will contain a section with
        this name. This allows you to keep the configuration file organized.
        If `None`, the settings will be stored outside of any section.
    """

    # TODO: Document which datatypes are supported

    # Any values from this class will be stored in the configuration file under
    # this section. This has to be set to a string. If empty, the values will be
    # set outside of any sections.
    section_name: t.ClassVar[str] = ""

    _rio_session_: session.Session | None = dataclasses.field(
        default=None, init=False, repr=False, compare=False
    )

    # Set of field names that have been modified and need to be saved
    _rio_dirty_attribute_names_: set[str] = dataclasses.field(
        default_factory=set, init=False, repr=False, compare=False
    )

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        if cls.section_name.startswith("section:"):
            raise ValueError(f"Section names may not start with 'section:'")

    @classmethod
    def _from_json(
        cls,
        settings_json: uniserde.JsonDoc,
        defaults: te.Self,
    ) -> te.Self:
        # Create the instance for this attachment. Bypass the constructor so the
        # instance doesn't immediately try to synchronize with the frontend.
        self = object.__new__(cls)
        settings_vars = vars(self)

        if cls.section_name:
            section = t.cast(
                dict[str, object],
                settings_json.get("section:" + cls.section_name, {}),
            )
        else:
            section = settings_json

        annotations = inspection.get_resolved_type_annotations(cls)

        for field_name, field_type in annotations.items():
            # Skip internal fields
            if field_name in UserSettings.__annotations__:
                continue

            # Try to parse the field value
            try:
                field_value = serialization.json_serde.from_json(
                    field_type,
                    section[field_name],
                )
            except (KeyError, uniserde.SerdeError):
                field_value = copy.deepcopy(getattr(defaults, field_name))

            # Set the field value
            settings_vars[field_name] = field_value

        return self

    # This function kinda ruins linting, so we'll hide it from the IDE
    def __setattr(self, name: str, value: t.Any) -> None:
        # These attributes doesn't exist yet during the constructor
        dct = vars(self)
        dirty_attribute_names = dct.setdefault(
            "_rio_dirty_attribute_names_", set()
        )

        # Set the attribute
        dct[name] = value

        # Ignore assignments to internal attributes
        if name in __class__.__annotations__:
            return

        # Mark it as dirty
        dirty_attribute_names.add(name)

        # Make sure a write back task is running
        # if self._rio_session_ is not None:
        #     self._rio_session_._save_settings_soon()

    if not t.TYPE_CHECKING:
        __setattr__ = __setattr

    def _equals(self, other: te.Self) -> bool:
        if type(self) != type(other):
            return False

        fields_to_compare = (
            all_class_fields(type(self)).keys()
            - all_class_fields(__class__).keys()
        )
        for field_name in fields_to_compare:
            if getattr(self, field_name) != getattr(other, field_name):
                return False

        return True



import asyncio
import json
import linecache
import logging
import signal
import socket
import sys
import threading
import time
import typing as t
from datetime import datetime, timedelta, timezone
from pathlib import Path

import revel
from revel import print

import rio.app_server.fastapi_server
import rio.arequests as arequests
import rio.cli
import rio.snippets

from ... import nice_traceback, project_config, utils, version
from ...debug.monkeypatches import apply_monkeypatches
from . import (
    app_loading,
    file_watcher_worker,
    run_models,
    uvicorn_worker,
    webview_worker,
)

# There is a hilarious problem with `watchfiles`: When a change occurs,
# `watchfiles` logs that change. If Python is configured to log into the project
# directory, then `watchfiles` will pick up on that new log line, thus
# triggering another change. Infinite loop, here we come!
#
# -> STFU, watchfiles!
logging.getLogger("watchfiles").setLevel(logging.CRITICAL + 1)


LOGO_TEXT = r"""  _____  _
 |  __ \(_)
 | |__) |_  ___
 |  _  /| |/ _ \
 | | \ \| | (_) |
 |_|  \_\_|\___/"""


class Arbiter:
    """
    Serves as a unified interface to running apps, coordinating the various
    workers necessary to do so.
    """

    def __init__(
        self,
        *,
        proj: project_config.RioProjectConfig,
        port: int | None,
        public: bool,
        quiet: bool,
        debug_mode: bool,
        run_in_window: bool,
        base_url: rio.URL | None,
    ) -> None:
        assert not (run_in_window and public), (
            "Can't run in a local window over the network, wtf!?"
        )

        self.proj = proj
        self.public = public
        self.quiet = quiet
        self.debug_mode = debug_mode
        self.run_in_window = run_in_window
        self.base_url = base_url

        # Workers communicate with the `RunningApp` by pushing events into this
        # queue.
        self._event_queue: asyncio.Queue[run_models.Event] = asyncio.Queue()

        # Contains the various workers, if running
        self._file_watcher_worker: (
            file_watcher_worker.FileWatcherWorker | None
        ) = None
        self._uvicorn_worker: uvicorn_worker.UvicornWorker | None = None
        self._webview_worker: webview_worker.WebViewWorker | None = None

        self._file_watcher_task: asyncio.Task[None] | None = None
        self._uvicorn_task: asyncio.Task[None] | None = None
        self._arbiter_task: asyncio.Task[None] | None = None

        # The theme to use for creating placeholder apps. This keeps the theme
        # consistent if for-example the user's app crashes and then a mock-app
        # is injected.
        self._app_theme: t.Union[
            rio.Theme,
            tuple[rio.Theme, rio.Theme],
        ] = rio.Theme.pair_from_colors()

        # Prefer to consistently run on the same port, as that makes it easier
        # to connect to - this way old browser tabs don't get invalidated
        # constantly.
        if port is None:
            if utils.port_is_free(self._host, 8000):
                self.port = 8000
            else:
                self.port = utils.choose_free_port(self._host)
        else:
            self.port = port

        # Used by the stop function for thread safety
        self._stopping_lock = threading.Lock()

        # This event resolves once the app is ready to be shown in the webview.
        self._server_is_ready = threading.Event()

        # Used to signal that the app should close
        self._stop_requested = threading.Event()

        # The mainloop used by the arbiter. This is set once asyncio is up and
        # running.
        self._mainloop: asyncio.AbstractEventLoop | None = None

    def push_event(self, event: run_models.Event) -> None:
        """
        Pushes an event into the event queue. Threadsafe.
        """
        assert self._mainloop is not None, (
            "Can't push events before asyncio is running"
        )
        rio.cli._logger.debug(f"Pushing arbiter event `{event}`")
        self._mainloop.call_soon_threadsafe(self._event_queue.put_nowait, event)

    @property
    def _host(self) -> str:
        return "0.0.0.0" if self.public else "127.0.0.1"

    @property
    def url(self) -> str:
        # If a base URL is specified, use that
        if self.base_url is not None:
            return str(self.base_url)

        # If running in public mode, use the device's IP on the local network.
        if self.public:
            # Get the local IP. This doesn't send data, because UDP is
            # connectionless.
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(("8.8.8.8", 80))
            local_ip = s.getsockname()[0]
            s.close()

        # Otherwise use localhost
        else:
            local_ip = "127.0.0.1"

        return f"http://{local_ip}:{self.port}"

    @property
    def running_tasks(self) -> t.Iterator[asyncio.Task[None]]:
        for task in (
            self._uvicorn_task,
            self._file_watcher_task,
            self._arbiter_task,
        ):
            if task is not None:
                yield task

    async def _fetch_published_rio_version_from_pypi(self) -> str:
        """
        Fetches the version number of the latest Rio available on PyPI, without
        any cache.

        ## Raises

        `ValueError`: If the version number could not be determined.

        """
        try:
            response = await arequests.request(
                "get",
                "https://pypi.org/pypi/rio-ui/json",
            )

            data = response.json()
            return data["info"]["version"]

        except (
            arequests.HttpError,
            json.JSONDecodeError,
        ) as exc:
            raise ValueError(f"Failed to fetch the latest Rio version: {exc}")

    async def get_published_rio_version(self) -> str:
        """
        Finds the version number of the latest Rio available on PyPI, using a
        cache to avoid unnecessary network requests.

        ## Raises

        `ValueError`: If the version number could not be determined.
        """
        # See if a cached version is available
        #
        # Asking the asset manager for a cache path will also ensure that the
        # cache directory exists.
        cache_path = utils.ASSET_MANAGER.get_cache_path(
            Path("rio-pypi-version.txt")
        )
        threshold = datetime.now(timezone.utc) - timedelta(hours=3)

        try:
            file_modified_at = cache_path.stat().st_mtime
            file_contents = cache_path.read_text()
        except OSError:
            pass
        else:
            if (
                datetime.fromtimestamp(file_modified_at, timezone.utc)
                >= threshold
            ):
                return file_contents.strip()

        # No cache exists. Fetch the version number from PyPI
        result = await self._fetch_published_rio_version_from_pypi()

        # Store the result in the cache
        try:
            cache_path.write_text(result)
        except OSError:
            pass

        return result

    def stop(self, *, keyboard_interrupt: bool) -> None:
        """
        Stops the app. This function can safely be called more than once, and at
        any point in time.
        """
        # Already shutting down? Then do nothing
        if not self._stopping_lock.acquire(blocking=False):
            return

        rio.cli._logger.debug("Stopping the app")

        # Visualize KeyboardInterrupts
        if keyboard_interrupt:
            print()
            print("[yellow]Interrupted[/]")
        else:
            print("Stopping")

        # Release anyone waiting for events
        self._stop_requested.set()
        self._server_is_ready.set()

        # Stop any running workers.
        #
        # The mainloop may not be running yet, if something went wrong before it
        # was started.
        def cancel_all_tasks() -> None:
            rio.cli._logger.debug(
                "Cancelling all arbiter tasks because the app is stopping"
            )

            for task in self.running_tasks:
                task.cancel("Arbiter is stopping")

        if self._mainloop is not None:
            self._mainloop.call_soon_threadsafe(cancel_all_tasks)

        # Stop the webview
        if self._webview_worker is not None:
            rio.cli._logger.debug(
                "Stopping the webview because the app is stopping"
            )
            self._webview_worker.request_stop()

    def try_load_app(
        self,
    ) -> tuple[
        rio.app_server.fastapi_server.FastapiServer,
        Exception | None,
    ]:
        """
        Tries to load the user's app. If it fails, a dummy app is created and
        returned, unless running in release mode. (In release mode screams and
        exits the entire process.)

        Returns the app server instance and an exception if the app could not be
        loaded.

        The app server is returned in case the user has called `as_fastapi` on
        their app instance. In that case the actual fastapi app should be
        hosted, so any custom routes take effect.
        """
        rio.cli._logger.debug("Trying to load the app")

        try:
            app_server = app_loading.load_user_app(self.proj)

        except app_loading.AppLoadError as err:
            # Announce the problem in the terminal
            # rio.cli._logger.exception(f"The app could not be loaded: {err}")
            revel.error(f"The app could not be loaded: {err}")

            if err.__cause__ is not None:
                nice_traceback.print_exception(
                    err.__cause__,
                    relpath=self.proj.project_directory,
                )

            # If running in release mode, no further attempts to load the app
            # will be made. This error is fatal.
            if not self.debug_mode:
                sys.exit(1)

            print()

            # Otherwise create a placeholder app which displays the error
            # message
            app = app_loading.make_error_message_app(
                err,
                self.proj.project_directory,
                self._app_theme,
            )
            app_server = app.as_fastapi()
            assert isinstance(
                app_server, rio.app_server.fastapi_server.FastapiServer
            )

            return app_server, err

        # Remember the app's theme. If in the future a placeholder app is used,
        # this theme will be used for it.
        self._app_theme = app_server.app._theme

        return app_server, None

    def run(self) -> None:
        assert not self._stop_requested.is_set()
        assert not self._server_is_ready.is_set()

        # Handle keyboard interrupts. KeyboardInterrupt exceptions are very
        # annoying to handle in async code, so instead we'll use a signal
        # handler.
        def keyboard_interrupt_handler(*_):
            self.stop(keyboard_interrupt=True)

            # Restore the original keyboard interrupt handler. Sometimes things
            # hang, and our graceful shutdown won't work. Restoring the original
            # handler gives the user a more powerful tool to work with.
            signal.signal(signal.SIGINT, original_keyboard_interrupt_handler)

        original_keyboard_interrupt_handler = signal.signal(
            signal.SIGINT, keyboard_interrupt_handler
        )

        # Do as much work as possible in asyncio land. If this crashes for any
        # reason, stop the world.
        async def asyncio_wrapper() -> None:
            try:
                await self._run_async()
            finally:
                self.stop(keyboard_interrupt=False)

        asyncio_thread = threading.Thread(
            target=lambda: asyncio.run(asyncio_wrapper()),
            # target=lambda: asyncio.run(self._run_async()),
            name="arbiter function of rio run",
        )
        asyncio_thread.start()

        # Wait for the http server to be ready
        rio.cli._logger.debug("Waiting for the http server to be ready")
        self._server_is_ready.wait()

        # Make sure the startup was successful
        if self._stop_requested.is_set():
            rio.cli._logger.debug(
                "Stopping the arbiter because a stop was requested before the startup sequence has finished"
            )
            asyncio_thread.join()
            return

        # The webview needs to be shown from the main thread. So, if running
        # inside of a window run the arbiter in a separate thread. Otherwise
        # just run it from this one.
        if self.run_in_window:
            self._webview_worker = webview_worker.WebViewWorker(
                push_event=self.push_event,
                debug_mode=self.debug_mode,
                url=self.url,
            )

            rio.cli._logger.debug("Starting the webview worker")
            assert self._uvicorn_worker is not None
            self._webview_worker.run(self._uvicorn_worker.app_server.app)

        # If not running in a webview, just wait
        else:
            rio.cli._logger.debug("Opening the browser")

            # Event.wait() blocks the SIGINT handler, so we must periodically
            # return to python land to react to keyboard interrupts.
            #
            # We could pass a timeout to `_stop_requested.wait()`, but
            # `time.sleep()` reacts to keyboard interrupts immediately, which is
            # preferable.
            while True:
                time.sleep(0.2)

                if self._stop_requested.is_set():
                    break

        # Make sure the main thread stays alive until the asyncio thread is
        # done. Otherwise we get weird errors like "Can't start thread during
        # interpreter shutdown" from asyncio.
        rio.cli._logger.debug(
            "The arbiter is waiting for the asyncio thread to finish"
        )
        asyncio_thread.join()

        rio.cli._logger.debug("Arbiter shutdown complete")

    async def _run_async(self) -> None:
        # Publish the task, so it can be cancelled
        self._arbiter_task = asyncio.current_task()

        # Make sure the webview module is available
        if self.run_in_window:
            try:
                from ... import webview_shim as webview_shim
            except ImportError:
                revel.fatal(
                    """The `window` extra is required to run apps inside of a window. Run `pip install "rio-ui[[window]"` to install it."""
                )

        # Make sure the app is cleanly shut down, even if the arbiter crashes
        # for whichever reason.
        #
        # This loop has a bad case of hiding important exceptions. It would
        # really be nice to only shut down on intentional errors and at least
        # log the others.
        try:
            await self._run_async_inner()

        # Keyboard interrupts are already handled by the signal handler
        except KeyboardInterrupt:
            pass

        # The worker is expected to be cancelled when the app is stopped
        except asyncio.CancelledError:
            pass

        # Anything else is an unintentional crash
        except Exception as err:
            revel.error("The arbiter has crashed.")
            revel.error("This is a bug in Rio - please report it")
            print()
            nice_traceback.print_exception(err)

            rio.cli._logger.exception("The arbiter has crashed")

            self.stop(keyboard_interrupt=False)

        finally:
            # Wait until all the other tasks are done. If we return, then
            # `asyncio.run` will cancel anything that's still running.
            rio.cli._logger.debug("Waiting for all arbiter tasks to finish")

            for task in self.running_tasks:
                if task is self._arbiter_task:
                    continue

                try:
                    await task
                except asyncio.CancelledError:
                    pass

            # Then cancel all remaining tasks to ensure the program exits
            rio.cli._logger.debug("Cancelling all remaining asyncio tasks")

            for task in asyncio.all_tasks():
                if task is self._arbiter_task:
                    continue

                task.cancel("Arbiter is shutting down")

    async def _run_async_inner(self) -> None:
        # Print some initial messages
        print()
        print(f"[bold green]{LOGO_TEXT}[/]")
        print()
        print()
        print("Starting...")

        # Expose the mainloop
        self._mainloop = asyncio.get_running_loop()

        # Make sure the chosen port is available
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            rio.cli._logger.debug(f"Checking if port {self.port} is available")

            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

            try:
                sock.bind((self._host, self.port))
            except OSError:
                revel.error(f"The port [bold]{self.port}[/] is already in use.")
                revel.error("Each port can only be used by one app at a time.")
                revel.error(
                    "Try using another port, or let Rio choose one for you, by not specifying any port."
                )
                self.stop(keyboard_interrupt=False)
                return

            # Fetch the most recent version of Rio. Since this needs to make
            # HTTP requests do it as early as possible
            rio_version_fetcher_task = asyncio.create_task(
                self.get_published_rio_version(),
                name="Fetch Rio version",
            )

            # If running in debug mode, install safeguards
            if self.debug_mode:
                rio.cli._logger.debug("Applying monkeypatches")
                apply_monkeypatches()

            # Try to load the app
            app_server, _ = self.try_load_app()

            # Start the file watcher
            if self.debug_mode:
                self._file_watcher_worker = (
                    file_watcher_worker.FileWatcherWorker(
                        push_event=self.push_event,
                        proj=self.proj,
                    )
                )
                self._file_watcher_task = asyncio.create_task(
                    self._file_watcher_worker.run(),
                    name="File watcher",
                )

            # Start the uvicorn worker
            uvicorn_is_ready_or_has_failed: asyncio.Future[None] = (
                asyncio.Future()
            )

            self._uvicorn_worker = uvicorn_worker.UvicornWorker(
                push_event=self.push_event,
                app_server=app_server,
                socket=sock,
                quiet=self.quiet,
                debug_mode=self.debug_mode,
                run_in_window=self.run_in_window,
                on_server_is_ready_or_failed=uvicorn_is_ready_or_has_failed,
                base_url=self.base_url,
            )

            self._uvicorn_task = asyncio.create_task(
                self._uvicorn_worker.run(),
                name="Uvicorn",
            )

            # Wait for the server to be ready or fails to start
            rio.cli._logger.debug("Waiting for uvicorn to be ready or to fail")
            await uvicorn_is_ready_or_has_failed

            # Let everyone else know that the server is ready
            rio.cli._logger.debug(
                "App startup complete and ready for connections"
            )
            self._server_is_ready.set()

            # The app has just successfully started. Inform the user that...
            #
            # ...their Rio version is outdated
            installed_rio_version = rio.__version__

            try:
                newest_rio_version = await rio_version_fetcher_task
            except ValueError as err:
                logging.warning(
                    f"Failed to fetch the newest Rio version. Skipping version check: {err}"
                )
            else:
                try:
                    installed_rio_version_parsed = version.Version.parse(
                        installed_rio_version
                    )
                except ValueError:
                    logging.warning(
                        f"Failed to parse installed Rio version: `{installed_rio_version}`"
                    )
                    installed_rio_version_parsed = version.Version(0)

                try:
                    newest_rio_version_parsed = version.Version.parse(
                        newest_rio_version
                    )
                except ValueError:
                    logging.warning(
                        f"Failed to parse newest Rio version: `{newest_rio_version}`"
                    )
                    newest_rio_version_parsed = version.Version(0)

                # From experience, people don't even notice the warnings in `rio
                # run` anymore after some time, because they show up so
                # frequently. Intentionally style this one differently, since
                # it's important and must be noticed.
                if newest_rio_version_parsed > installed_rio_version_parsed:
                    revel.print(
                        f"[bg-yellow] [/] [bold yellow]Rio [bold]{newest_rio_version}[/] is available![/]"
                    )
                    revel.print(
                        "[bg-yellow] [/] [bold yellow]Please update to get the newest features, bugfixes and security updates.[/]"
                    )
                    revel.print(
                        "[bg-yellow] [/] Run `[bold]pip install --upgrade rio-ui[/]` to get the newest version."
                    )
                    print()

            # ...debug mode is enabled
            if self.debug_mode:
                revel.warning("Debug mode is enabled.")
                revel.warning(
                    "Debug mode includes helpful tools for development, but is slower and disables some safety checks. Never use it in production!"
                )
                revel.warning("Run with `--release` to disable debug mode.")
                print()

            # ...public mode is enabled (or not)
            if self.public:
                revel.warning(
                    "Running in public mode. All devices on your network can access the app."
                )
                revel.warning(
                    "Only run in public mode if you trust your network!"
                )
                revel.warning(
                    "Run without `--public` to limit access to this device."
                )
                print()
            elif not self.run_in_window:
                print(
                    "[dim]Running in [/]local[dim] mode. Only this device can access the app.[/]"
                )
                print()

            # ...where the app is running
            if not self.run_in_window:
                revel.success(f"The app is running at [bold]{self.url}[/]")
                print()

            # Support us, pretty please?
            try:
                print(
                    "🗨  Join other Developers on Discord [dim]—[/] https://discord.gg/7ejXaPwhyH"
                )
                print(
                    "✨ Star Rio on GitHub [dim]—[/] https://github.com/rio-labs/rio"
                )
                print(
                    "🐞 Report a bug [dim]—[/] https://github.com/rio-labs/rio/issues/new/choose"
                )
                print("📣 Spread the word")
            except UnicodeEncodeError:
                pass

            # Keep track when the app was last reloaded
            last_reload_started_at = -1

            # Listen for events and react to them
            while True:
                event = await self._event_queue.get()
                rio.cli._logger.debug(f"Received arbiter event `{event}`")

                # A file has changed
                if isinstance(event, run_models.FileChanged):
                    # Ignore events that happened before the last reload started
                    if event.timestamp_nanoseconds < last_reload_started_at:
                        continue

                    # Display to the user that a reload is happening
                    rel_path = event.path_to_file.relative_to(
                        self.proj.project_directory
                    )
                    print()
                    print(f"[bold]{rel_path}[/] has changed -> Reloading")

                    # Update the timestamp
                    last_reload_started_at = time.monotonic_ns()

                    # If the `rio.toml` has changed, reload the project
                    # configuration in addition to the Python app
                    if event.path_to_file == self.proj.rio_toml_path:
                        msg = "Reloading project configuration from `rio.toml`"
                        rio.cli._logger.info(msg)
                        print(msg)
                        self._try_reload_project_config()

                    # Restart the app
                    await self._restart_app()

                # Somebody requested that the app should stop
                elif isinstance(event, run_models.StopRequested):
                    self.stop(keyboard_interrupt=False)
                    break

                # ???
                else:
                    raise NotImplementedError(f'Unknown event "{event}"')

    def _try_reload_project_config(self) -> None:
        """
        Attempts to reload the project's configuration from disk. Should this
        fail, details are logged & printed, but no exception is raised. Better
        to have the app running in an outdated state than crash.
        """
        # Try to reload the project configuration
        try:
            self.proj.discard_changes_and_reload()

        # Oh uh.
        except OSError as err:
            msg = f"Couldn't reload the project configuration: {err}"
            revel.error(msg)
            rio.cli._logger.exception(msg)

        except rio.InvalidProjectConfigError as err:
            msg = f"The project configuration is invalid: {err}"
            revel.error(msg)
            rio.cli._logger.error(msg)

    def _spawn_traceback_popups(self, err: t.Union[str, BaseException]) -> None:
        """
        Displays a popup with the traceback in the rio UI.
        """
        assert self._uvicorn_worker is not None
        assert self._uvicorn_worker.app_server is not None

        rio.cli._logger.debug("Spawning traceback popups")

        popup_html = app_loading.make_traceback_html(
            err=err,
            project_directory=self.proj.project_directory,
        )

        for (
            session
        ) in self._uvicorn_worker.app_server._active_session_tokens.values():
            self._evaluate_javascript_in_session_if_connected(
                session,
                f"""
// Override the popup with the traceback message
let popup = document.querySelector(".rio-connection-lost-popup-container");
popup.innerHTML = {json.dumps(popup_html)};

// Spawn the popup
window.setConnectionLostPopupVisible(true);
""",
            )

    async def _restart_app(self) -> None:
        assert self._uvicorn_worker is not None

        rio.cli._logger.debug("Beginning app restart sequence")

        app_server = self._uvicorn_worker.app_server
        assert app_server is not None

        # To makes sure that clients can't (re-)connect while we're in the
        # middle of reloading, tell the app server not to accept new websocket
        # connections
        with app_server.temporarily_disable_new_session_creation():
            # Close all open sessions. It's important that `on_session_close` is
            # executed *before* we reload the module. (And before `on_app_close`
            # is called.)
            #
            # We'll save a list of open sessions so that we can re-create them
            # later.
            sessions = app_server._active_session_tokens.values()

            await asyncio.gather(
                *[
                    session._close(close_remote_session=False)
                    for session in sessions
                ]
            )

            # Call `on_app_close`. This happens automatically when the app
            # server shuts down, but since we're just swapping out the app, we
            # have to do it manually.
            await app_server._call_on_app_close()

        # Clear the linecache. This is used to fetch code for tracebacks; if we
        # don't clear the cache and multiple errors happen in a row, then the
        # tracebacks will display outdated code.
        linecache.checkcache()

        # Load the user's app again
        new_app_server, loading_error = self.try_load_app()

        with new_app_server.temporarily_disable_new_session_creation():
            # Replace the app which is currently hosted by uvicorn
            self._uvicorn_worker.replace_app(new_app_server)

            if self._webview_worker is not None:
                self._webview_worker.update_window_for_app(new_app_server.app)

            # The app has changed, but the uvicorn server is still the same.
            # Because of this, uvicorn won't call the `on_app_start` function -
            # do it manually.
            await new_app_server._call_on_app_starts()

            # There is a subtlety here. Sessions which have requested their
            # index.html, but aren't yet connected to the websocket cannot
            # receive messages. So `_spawn_traceback_popups` will skip them.
            # This is fine as long as `self._app_server.app` has already been
            # assigned, since this ensures that any new websocket connections
            # will receive the new app anyway.
            #
            # -> This MUST happen after the new app has already been injected
            if loading_error is not None:
                self._spawn_traceback_popups(loading_error)

        revel.success("Ready")

    def _evaluate_javascript_in_session_if_connected(
        self,
        session: rio.Session,
        javascript_source: str,
    ) -> None:
        """
        Runs the given javascript source in the given session. Does not
        wait for or return the result.

        Does nothing if the session hasn't fully connected yet.
        """
        # If this session isn't done connecting, just return
        if not session._is_connected:
            return

        # Run the javascript in a task
        async def evaljs_as_coroutine() -> None:
            await session._evaluate_javascript(javascript_source)

        session.create_task(
            evaljs_as_coroutine(),
            name=f"Eval JS in session {session}",
        )



from __future__ import annotations

# <additional-imports>
import rio

from .. import components as comps
from .. import data_models, theme

# </additional-imports>


# <component>
class BlogHeader(rio.Component):
    """
    A component to display the header of a blog post.


    ## Attributes:

    `blog_post`: The data for the blog post, including category, title,
    description, date, and author details.
    """

    blog_post: data_models.BlogPost

    def build(self) -> rio.Component:
        """
        Build the BlogHeader component.

        The layout includes:
        - A row with the category button, separator, and publication date.
        - A prominently styled title.
        - A description text.
        - An author card with metadata.
        - A separator line for distinction.
        """
        return rio.Column(
            # Row with category button, separator, and publication date
            rio.Row(
                comps.OutlinedColoredButton(
                    text=self.blog_post.category,
                    corner_radius=self.session.theme.corner_radius_small,
                    font_size=0.8,
                    font_weight="normal",
                    margin_x_row=0.5,
                    margin_y_row=0.3,
                    align_x=0,
                ),
                rio.Text(
                    "·",
                    style=theme.DARKER_TEXT,
                ),
                rio.Text(
                    self.blog_post.date,
                    style=theme.DARK_TEXT_SMALLER,
                ),
                spacing=0.5,  # Space between elements in the row
                align_x=0,  # Align row content to the left
                align_y=0,  # Align row vertically to the top
            ),
            # Blog post title
            rio.Text(
                self.blog_post.title,
                overflow="wrap",
                style=rio.TextStyle(
                    fill=theme.TEXT_FILL_SLIGHTLY_BRIGHTER,
                    font_weight="bold",
                    font_size=2,
                ),
            ),
            # Blog post description
            rio.Text(
                self.blog_post.description,
                overflow="wrap",
                style=rio.TextStyle(
                    fill=theme.TEXT_FILL_DARKER,
                    font_size=1.2,
                ),
            ),
            # Author card with details
            comps.AuthorCard(blog_post=self.blog_post),
            # Separator line for distinction
            rio.Separator(color=self.session.theme.neutral_color),
            spacing=1,  # Space between elements in the column
        )


# </component>



from __future__ import annotations

# <additional-imports>
import rio

from ... import components as comps
from ... import data_models, utils

# </additional-imports>


# <component>
# List of image URLs to be used in the AdventureInAlps page
image_url: list[str] = [
    "https://images.pexels.com/photos/5366526/pexels-photo-5366526.jpeg?auto=compress&cs=tinysrgb&w=1260&h=750&dpr=1",
    "https://images.pexels.com/photos/2724664/pexels-photo-2724664.jpeg?auto=compress&cs=tinysrgb&w=1260&h=750&dpr=1",
    "https://images.pexels.com/photos/4969837/pexels-photo-4969837.jpeg?auto=compress&cs=tinysrgb&w=1260&h=750&dpr=1",
    "https://images.pexels.com/photos/24604765/pexels-photo-24604765/free-photo-of-wanderer-steht-am-see-und-blickt-auf-das-matterhorn.jpeg?auto=compress&cs=tinysrgb&w=1260&h=750&dpr=1",
    "https://images.pexels.com/photos/2444429/pexels-photo-2444429.jpeg?auto=compress&cs=tinysrgb&w=1260&h=750&dpr=1",
]


@rio.page(
    name="adventure",
    url_segment="adventure-in-the-swiss-alps",
)
class AdventureInAlps(rio.Component):
    """
    AdventureInAlps Page Component.

    This page showcases an adventure in the Swiss Alps, including a "detailed
    descriptions" and images to engage users.
    """

    def build(self) -> rio.Component:
        # Retrieve the device type from the session data
        device = self.session[data_models.PageLayout].device

        # get the URL name of the active page
        url_name = self.session.active_page_url.name
        # Get single blog post by URL key
        blog_post = utils.BLOG_POSTS_BY_URL["/" + url_name]

        # Define layout parameters based on device type
        if device == "desktop":
            min_img_width = 80  # Fixed image width for desktop
            min_img_height = 40  # Fixed image height for desktop
            min_img_height_row = (
                20  # Fixed image height within a row for desktop
            )
            spacing = 2  # Spacing between images for desktop

        else:
            margin_x = 1
            # Calculate the minimum image width for mobile based on the window
            # width and margin
            min_img_width = self.session.window_width - margin_x * 2
            min_img_height = 13  # Fixed image height for mobile
            min_img_height_row = 7  # Fixed image height within a row for mobile
            spacing = 0.5  # Spacing between images for mobile

        # Calculate image width within a row based on spacing and number of
        # images
        min_img_width_row = (min_img_width - 3 * spacing) / 4

        # Create a row to hold multiple images
        image_content = rio.Row(spacing=spacing)
        # Ignore the first image in the list as it is displayed separately
        for i in range(1, len(image_url)):
            image_content.add(
                rio.Image(
                    rio.URL(image_url[i]),
                    fill_mode="stretch",
                    corner_radius=self.session.theme.corner_radius_medium,
                    min_width=min_img_width_row,
                    min_height=min_img_height_row,
                )
            )

        # Construct the main content column
        content = rio.Column(
            comps.BlogHeader(blog_post),
            rio.Markdown(
                """
# Lorem ipsum dolor sit amet, consetetur sadipscing elitr

Duis autem vel eum iriure dolor in hendrerit in vulputate velit esse molestie
consequat, vel illum dolore eu feugiat nulla facilisis at vero eros et accumsan
et iusto odio dignissim qui blandit praesent luptatum zzril delenit augue duis
dolore te feugait nulla facilisi. Lorem ipsum dolor sit amet, consectetuer
adipiscing elit, sed diam nonummy nibh euismod tincidunt ut laoreet dolore magna
aliquam erat volutpat. 

Ut wisi enim ad minim veniam, quis nostrud exerci tation ullamcorper suscipit
lobortis nisl ut aliquip ex ea commodo consequat. Duis autem vel eum iriure
dolor in hendrerit in vulputate velit esse molestie consequat, vel illum dolore
eu feugiat nulla facilisis at vero eros et accumsan et iusto odio dignissim qui
blandit praesent luptatum zzril delenit augue duis dolore te feugait nulla
facilisi. 
"""
            ),
            # Add the first image separately with a larger size
            rio.Image(
                rio.URL(image_url[0]),
                fill_mode="stretch",
                corner_radius=self.session.theme.corner_radius_medium,
                min_width=min_img_width,
                min_height=min_img_height,
            ),
            rio.Markdown(
                """
## Lorem ipsum dolor sit amet, consetetur sadipscing elitr

Nam liber tempor cum soluta nobis eleifend option congue nihil imperdiet doming
id quod mazim placerat facer possim assum. Lorem ipsum dolor sit amet,
consectetuer adipiscing elit, sed diam nonummy nibh euismod tincidunt ut laoreet
dolore magna aliquam erat volutpat. Ut wisi enim ad minim veniam, quis nostrud
exerci tation ullamcorper suscipit lobortis nisl ut aliquip ex ea commodo
consequat. 

At vero eos et accusam et justo duo dolores et ea rebum. Stet clita kasd
gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet. Lorem ipsum
dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor
invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. At vero eos
et accusam et justo duo dolores et ea rebum. Stet clita kasd gubergren, no sea
takimata sanctus est Lorem ipsum dolor sit amet. Lorem ipsum dolor sit amet,
consetetur sadipscing elitr, At accusam aliquyam diam diam dolore dolores duo
eirmod eos erat, et nonumy sed tempor et et invidunt justo labore Stet clita ea
et gubergren, kasd magna no rebum. sanctus sea sed takimata ut vero voluptua.
est Lorem ipsum dolor sit amet. Lorem ipsum dolor sit amet, consetetur
sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore
magna aliquyam erat. 
"""
            ),
            # Add the image row content
            image_content,
            rio.Markdown(
                """
## Lorem ipsum dolor sit amet, consetetur sadipscing elitr

Consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et
dolore magna aliquyam erat, sed diam voluptua. At vero eos et accusam et justo
duo dolores et ea rebum. Stet clita kasd gubergren, no sea takimata sanctus est
Lorem ipsum dolor sit amet. Lorem ipsum dolor sit amet, consetetur sadipscing
elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam
erat, sed diam voluptua. At vero eos et accusam et justo duo dolores et ea
rebum. Stet clita kasd gubergren, no sea takimata sanctus est Lorem ipsum dolor
sit amet. Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam
nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam erat, sed diam
voluptua. At vero eos et accusam et justo duo dolores et ea rebum. Stet clita
kasd gubergren, no sea takimata sanctus. 

Duis autem vel eum iriure dolor in hendrerit in vulputate velit esse molestie
consequat, vel illum dolore eu feugiat nulla facilisis. 
"""
            ),
            spacing=2,
        )

        # Return the appropriate container based on device type
        if device == "desktop":
            return rio.Container(
                content=content,
                align_x=0.5,  # Center the container horizontally
                min_width=80,  # Fixed width for desktop layout
            )
        else:
            return rio.Container(
                content=content,
                margin_x=1,  # Horizontal margin for mobile layout
            )


# </component>



from __future__ import annotations

import functools
import json
import sys
import typing as t

import PIL.Image
import PIL.ImageDraw
import PIL.ImageFont

import rio
import rio.components.fundamental_component
import rio.components.root_components
from rio.data_models import UnittestComponentLayout

from .. import serialization

R = t.TypeVar("R")
P = t.ParamSpec("P")


# These components pass on the entirety of the available space to their
# children
FULL_SIZE_SINGLE_CONTAINERS: set[type[rio.Component]] = {
    rio.Card,
    rio.CustomListItem,
    rio.KeyEventListener,
    rio.Link,
    rio.MouseEventListener,
    rio.Rectangle,
    rio.Slideshow,
    rio.Stack,
    rio.Switcher,
}


def specialized(func: t.Callable[P, R]) -> t.Callable[P, R]:
    """
    Decorator that defers to a specialized method if one exists. If none is
    found, calls the original method.

    For example, consider a method called `foo`. If the function is called with
    a `rio.Row` instance as `self`, this will call `foo_Row` if it exists, and
    `foo` otherwise.
    """

    def result(self, component, *args, **kwargs) -> t.Any:
        # First, check if a specialized method for this component exists
        try:
            method = getattr(
                self,
                f"{func.__name__}_{type(component).__name__}",
            )
        except AttributeError:
            # Special case: A lot of containers behave in the same way - they pass
            # on all space. Avoid having to implement them all separately.
            if type(component) in FULL_SIZE_SINGLE_CONTAINERS or not isinstance(
                component,
                rio.components.fundamental_component.FundamentalComponent,
            ):
                method = getattr(self, f"{func.__name__}_SingleContainer")
            else:
                # If all else fails, use the default implementation
                method = functools.partial(func, self)

        return method(component, *args, **kwargs)  # type: ignore (wtf?)

    return result  # type: ignore


def _linear_container_get_major_axis_natural_size(
    child_requested_sizes: list[float],
    spacing: float,
    proportions: None | t.Literal["homogeneous"] | t.Sequence[float],
) -> float:
    # Spacing
    result = spacing * (len(child_requested_sizes) - 1)

    # Children
    #
    # No proportions
    if proportions is None:
        for child_requested_size in child_requested_sizes:
            result += child_requested_size

    # Proportions
    else:
        if proportions == "homogeneous":
            proportions = [1] * len(child_requested_sizes)
        else:
            proportions = list(proportions)

        # Find the width of 1 unit of proportions
        width_per_proportion = 0

        for child_size, proportion in zip(child_requested_sizes, proportions):
            width_per_proportion = max(
                width_per_proportion,
                child_size / proportion,
            )

        # Request the correct amount of space
        result += width_per_proportion * sum(proportions)

    return result


def _linear_container_get_major_axis_allocated_sizes(
    container_allocated_size: float,
    child_requested_sizes: list[float],
    child_growers: list[bool],
    spacing: float,
    proportions: None | t.Literal["homogeneous"] | t.Sequence[float],
) -> list[tuple[float, float]]:
    starts_and_sizes: list[tuple[float, float]] = []

    # Allow the code below to assume there is at least one child
    if not child_requested_sizes:
        return []

    # No proportions
    if proportions is None:
        cur_x = 0

        # Prepare for superfluous space
        additional_space = (
            container_allocated_size
            - sum(child_requested_sizes)
            - spacing * (len(child_requested_sizes) - 1)
        )

        n_growers = sum(child_growers)

        if n_growers == 0:
            additional_space_per_component = additional_space / len(
                child_requested_sizes
            )
            additional_space_per_grower = 0
        else:
            additional_space_per_component = 0
            additional_space_per_grower = additional_space / n_growers

        for child_requested_size, child_grow in zip(
            child_requested_sizes, child_growers
        ):
            # Determine how much space to pass on
            child_allocated_size = (
                child_requested_size + additional_space_per_component
            )

            if child_grow:
                child_allocated_size += additional_space_per_grower

            # Store the result
            starts_and_sizes.append((cur_x, child_allocated_size))
            cur_x += child_allocated_size + spacing

    # Proportions
    else:
        if proportions == "homogeneous":
            proportions = [1] * len(child_requested_sizes)
        else:
            proportions = list(proportions)

        # Find the width of 1 unit of proportions
        available_space = container_allocated_size - spacing * (
            len(child_requested_sizes) - 1
        )
        width_per_proportion = available_space / sum(proportions)

        # Pass on the correct amount of space
        cur_x = 0

        for proportion in proportions:
            starts_and_sizes.append((cur_x, width_per_proportion * proportion))
            cur_x += width_per_proportion * proportion + spacing

    # Done
    return starts_and_sizes


def calculate_alignment(
    allocated_outer_size: float,
    requested_inner_size: float,
    margin_start: float,
    margin_end: float,
    align: float | None,
) -> tuple[float, float]:
    # If no alignment is specified pass on all space
    if align is None:
        return margin_start, allocated_outer_size - margin_start - margin_end

    # If a margin is specified, only pass on the minimum amount of space and
    # distribute superfluous space
    additional_space = (
        allocated_outer_size - requested_inner_size - margin_start - margin_end
    )
    return margin_start + additional_space * align, requested_inner_size


def iter_direct_tree_children(
    component: rio.Component,
) -> t.Iterable[rio.Component]:
    """
    Iterates over the children of a component. In particular, the children which
    are part of the component tree, rather than those stored in the components
    attributes.
    """

    # Fundamental components can have any number of children
    if isinstance(
        component,
        rio.components.fundamental_component.FundamentalComponent,
    ):
        yield from component._iter_referenced_components_()

    # High level components have a single child: their build result
    else:
        yield component._build_data_.build_result  # type: ignore


class Layouter:
    session: rio.Session

    window_width: float
    window_height: float

    # Map components to their layouts. One for client-side, one for server-side.
    _layouts_should: dict[int, UnittestComponentLayout]
    _layouts_are: dict[int, UnittestComponentLayout]

    # Function to filter uninteresting components. If this returns `False` for a
    # given component, that component and all of its children are ignored.
    _filter: t.Callable[[rio.Component], bool]

    # All components in the session, ordered such that each parent appears
    # before its children.
    _ordered_components: list[rio.Component]

    # Construction of this class is asynchronous. Make sure nobody does anything
    # silly.
    def __init__(self) -> None:
        raise TypeError(
            "Creating this class is asynchronous. Use the `create` method instead."
        )

    @staticmethod
    async def create(
        session: rio.Session,
        *,
        filter: t.Callable[[rio.Component], bool] = lambda _: True,
    ) -> Layouter:
        # Create a new instance
        self = Layouter.__new__(Layouter)
        self.session = session
        self._filter = filter

        # Get a sorted list of components. Each parent appears before its
        # children.
        self._ordered_components = list(
            self._get_toposorted(self.session._high_level_root_component)
        )

        # Fetch client-side information. This includes things such as the window
        # size and the natural size of components.
        #
        # While it is obviously the job of this class to calculate the layout,
        # there is little reason to re-implement the sizing logic of every
        # single component. Instead, fetch the natural sizes of leaf components,
        # and then make a pass over all parent components to determine the
        # correct layout.
        client_info = await self.session._get_unittest_client_layout_info()

        self.window_width = client_info.window_width
        self.window_height = client_info.window_height
        self._layouts_are = client_info.component_layouts

        # Make sure the received components match expectations
        component_ids_client = set(self._layouts_are.keys())
        component_ids_server = {c._id_ for c in self._ordered_components}

        missing_component_ids = component_ids_server - component_ids_client
        assert not missing_component_ids, missing_component_ids

        superfluous_component_ids = component_ids_client - component_ids_server
        assert not superfluous_component_ids, superfluous_component_ids

        # Compute server-side layouts
        self._layouts_should = {}
        self._compute_layouts_should(
            ordered_components=self._ordered_components,
        )

        # Done!
        return self

    def get_component_by_id(self, component_id: int) -> rio.Component:
        return self.session._weak_components_by_id[component_id]

    def get_layout_by_key(
        self,
        key: str | int,
    ) -> UnittestComponentLayout:
        """
        Returns the layout for the component with the given key.

        This function is intended to be used for additional tests after one has
        already checked that all attributes from layout-should and layout-are
        match. Because of this, this function simply returns the "should"
        variant of the layout, since it is bound to be identical to the other
        one anyway.

        However, the "should" variant has one advantage: Because it wasn't
        calculated by a browser with real-world pixels, any contained values are
        exact, meaning you can compare the values using a simple `==` rather
        than a fuzzy comparison.


        ## Raises

        `KeyError`: If there is no component with the given key.
        """

        for component in self._ordered_components:
            if component.key == key:
                return self._layouts_should[component._id_]

        raise KeyError(f"There is no component with key `{key}`")

    def _get_toposorted(
        self,
        root: rio.Component,
    ) -> t.Iterable[rio.Component]:
        """
        Returns the component, as well as all direct and indirect children. The
        results are ordered such that each parent appears before its children.
        """

        to_do: list[rio.Component] = [root]

        while to_do:
            # Track this component
            current = to_do.pop()
            yield current

            # Recur into children
            to_do.extend(iter_direct_tree_children(current))

    def _compute_layouts_should(
        self,
        ordered_components: list[rio.Component],
    ) -> None:
        # Pre-allocate layouts for each component. Also, ...
        #
        # 1. Update natural & requested width
        for component in reversed(ordered_components):
            layout_is = self._layouts_are[component._id_]
            layout_should = UnittestComponentLayout(
                natural_width=-1,
                natural_height=-1,
                requested_inner_width=-1,
                requested_inner_height=-1,
                requested_outer_width=-1,
                requested_outer_height=-1,
                allocated_outer_width=-1,
                allocated_outer_height=-1,
                allocated_inner_width=-1,
                allocated_inner_height=-1,
                left_in_viewport_outer=-1,
                top_in_viewport_outer=-1,
                left_in_viewport_inner=-1,
                top_in_viewport_inner=-1,
                parent_id=layout_is.parent_id,
                aux={},
            )
            self._layouts_should[component._id_] = layout_should

            # Let the component update its natural width
            self._update_natural_width(component)
            assert layout_should.natural_width >= -0.1, layout_should

            # Derive the requested width
            min_width = (
                component.min_width
                if isinstance(component.min_width, (int, float))
                else 0
            )
            layout_should.requested_inner_width = max(
                layout_should.natural_width, min_width
            )

            # Account for rounding errors
            layout_should.requested_outer_width = (
                layout_should.requested_inner_width
                + component._effective_margin_left_
                + component._effective_margin_right_
            )

        # 2. Update allocated width
        for component in ordered_components:
            layout = self._layouts_should[component._id_]

            left, width = calculate_alignment(
                allocated_outer_size=layout.allocated_outer_width,
                requested_inner_size=layout.requested_inner_width,
                margin_start=component._effective_margin_left_,
                margin_end=component._effective_margin_right_,
                align=component.align_x,
            )
            layout.left_in_viewport_inner = layout.left_in_viewport_outer + left
            layout.allocated_inner_width = width

            self._update_allocated_width(component)

        # 3. Update natural & requested height
        for component in reversed(ordered_components):
            layout_should = self._layouts_should[component._id_]

            # Let the component update its natural height
            self._update_natural_height(component)
            assert layout_should.natural_height >= -0.1, layout_should

            # Derive the requested height
            min_height = (
                component.min_height
                if isinstance(component.min_height, (int, float))
                else 0
            )
            layout_should.requested_inner_height = max(
                layout_should.natural_height, min_height
            )

            layout_should.requested_outer_height = (
                layout_should.requested_inner_height
                + component._effective_margin_top_
                + component._effective_margin_bottom_
            )

        # 4. Update allocated height
        for component in ordered_components:
            layout = self._layouts_should[component._id_]

            top, height = calculate_alignment(
                allocated_outer_size=layout.allocated_outer_height,
                requested_inner_size=layout.requested_inner_height,
                margin_start=component._effective_margin_top_,
                margin_end=component._effective_margin_bottom_,
                align=component.align_y,
            )
            layout.top_in_viewport_inner = layout.top_in_viewport_outer + top
            layout.allocated_inner_height = height

            self._update_allocated_height(component)

    @specialized
    def _update_natural_width(
        self,
        component: rio.Component,
    ) -> None:
        """
        Updates the natural width for the given component. This assumes that all
        children already have their requested width set.
        """

        # Default implementation: Trust the client
        layout_should = self._layouts_should[component._id_]
        layout_is = self._layouts_are[component._id_]

        layout_should.natural_width = layout_is.natural_width

    def _update_natural_width_Row(
        self,
        component: rio.Row,
    ) -> None:
        # Prepare
        layout = self._layouts_should[component._id_]

        child_widths: list[float] = []

        for child in component._iter_referenced_components_():
            child_layout = self._layouts_should[child._id_]
            child_widths.append(child_layout.requested_outer_width)

        # Update
        layout.natural_width = _linear_container_get_major_axis_natural_size(
            child_requested_sizes=child_widths,
            spacing=component.spacing,
            proportions=component.proportions,
        )

    def _update_natural_width_Column(
        self,
        component: rio.Column,
    ) -> None:
        # Max of all Children
        layout = self._layouts_should[component._id_]
        layout.natural_width = 0

        for child in component._iter_referenced_components_():
            child_layout = self._layouts_should[child._id_]
            layout.natural_width = max(
                layout.natural_width, child_layout.requested_outer_width
            )

    def _update_natural_width_Overlay(
        self,
        component: rio.Overlay,
    ) -> None:
        layout = self._layouts_should[component._id_]
        layout.natural_width = 0

    def _update_natural_width_SingleContainer(
        self,
        component: rio.Component,
    ) -> None:
        # Prepare
        layout = self._layouts_should[component._id_]
        layout.natural_width = 0

        # Pass on all space
        for child in iter_direct_tree_children(component):
            child_layout = self._layouts_should[child._id_]
            layout.natural_width = max(
                layout.natural_width,
                child_layout.requested_outer_width,
            )

    @specialized
    def _update_allocated_width(
        self,
        component: rio.Component,
    ) -> None:
        """
        Updates the allocated width of all of the component's children. This
        assumes that the component itself already has its requested width set.

        Furthermore, this also assigns the `left_in_viewport` attribute of all
        children.
        """
        # Default implementation: Trust the client
        for child in iter_direct_tree_children(component):
            child_layout_should = self._layouts_should[child._id_]
            child_layout_is = self._layouts_are[child._id_]

            child_layout_should.left_in_viewport_outer = (
                child_layout_is.left_in_viewport_outer
            )

            child_layout_should.allocated_outer_width = (
                child_layout_is.allocated_outer_width
            )

    def _update_allocated_width_HighLevelRootComponent(
        self,
        component: rio.components.root_components.HighLevelRootComponent,
    ) -> None:
        # Since the HighLevelRootComponent doesn't have a parent, it has to set
        # its own allocation
        layout_should = self._layouts_should[component._id_]
        layout_is = self._layouts_are[component._id_]

        # Because scrolling differs between debug mode and release mode (user
        # content scrolls vs browser scrolls), we'll just copy the values from
        # the client.
        layout_should.left_in_viewport_outer = layout_is.left_in_viewport_outer
        layout_should.left_in_viewport_inner = layout_is.left_in_viewport_inner

        layout_should.allocated_outer_width = layout_is.allocated_outer_width
        layout_should.allocated_inner_width = layout_is.allocated_inner_width

        # Then behave like a regular SingleContainer
        self._update_allocated_width_SingleContainer(component)

    def _update_allocated_width_Row(
        self,
        component: rio.Row,
    ) -> None:
        # Prepare
        layout = self._layouts_should[component._id_]

        child_widths: list[float] = []

        for child in component._iter_referenced_components_():
            child_layout = self._layouts_should[child._id_]
            child_widths.append(child_layout.requested_outer_width)

        # Update
        starts_and_sizes = _linear_container_get_major_axis_allocated_sizes(
            container_allocated_size=layout.allocated_inner_width,
            child_requested_sizes=child_widths,
            child_growers=[
                child.grow_x
                for child in component._iter_referenced_components_()
            ],
            spacing=component.spacing,
            proportions=component.proportions,
        )

        for child, (left, width) in zip(
            component._iter_referenced_components_(), starts_and_sizes
        ):
            child_layout = self._layouts_should[child._id_]
            child_layout.left_in_viewport_outer = (
                layout.left_in_viewport_inner + left
            )
            child_layout.allocated_outer_width = width

    def _update_allocated_width_Column(
        self,
        component: rio.Column,
    ) -> None:
        layout = self._layouts_should[component._id_]

        for child in component._iter_referenced_components_():
            child_layout = self._layouts_should[child._id_]
            child_layout.left_in_viewport_outer = layout.left_in_viewport_inner
            child_layout.allocated_outer_width = layout.allocated_inner_width

    def _update_allocated_width_Overlay(
        self,
        component: rio.Overlay,
    ) -> None:
        child_layout = self._layouts_should[component.content._id_]
        child_layout.left_in_viewport_outer = 0
        child_layout.allocated_outer_width = self.window_width

    def _update_allocated_width_SingleContainer(
        self,
        component: rio.Component,
    ) -> None:
        # Prepare
        layout = self._layouts_should[component._id_]

        # Pass on all space
        for child in iter_direct_tree_children(component):
            child_layout = self._layouts_should[child._id_]
            child_layout.left_in_viewport_outer = layout.left_in_viewport_inner
            child_layout.allocated_outer_width = layout.allocated_inner_width

    @specialized
    def _update_natural_height(
        self,
        component: rio.Component,
    ) -> None:
        """
        Updates the natural height for the given component. This assumes that
        all children already have their requested height set.
        """
        # Default implementation: Trust the client
        layout_should = self._layouts_should[component._id_]
        layout_is = self._layouts_are[component._id_]

        layout_should.natural_height = layout_is.natural_height

    def _update_natural_height_Row(
        self,
        component: rio.Row,
    ) -> None:
        # Max of all Children
        layout = self._layouts_should[component._id_]
        layout.natural_height = 0

        for child in component._iter_referenced_components_():
            child_layout = self._layouts_should[child._id_]
            layout.natural_height = max(
                layout.natural_height, child_layout.requested_outer_height
            )

    def _update_natural_height_Column(
        self,
        component: rio.Column,
    ) -> None:
        # Prepare
        layout = self._layouts_should[component._id_]

        child_heights: list[float] = []

        for child in component._iter_referenced_components_():
            child_layout = self._layouts_should[child._id_]
            child_heights.append(child_layout.requested_outer_height)

        # Update
        layout.natural_height = _linear_container_get_major_axis_natural_size(
            child_requested_sizes=child_heights,
            spacing=component.spacing,
            proportions=component.proportions,
        )

    def _update_natural_height_Overlay(
        self,
        component: rio.Overlay,
    ) -> None:
        layout = self._layouts_should[component._id_]
        layout.natural_height = 0

    def _update_natural_height_SingleContainer(
        self,
        component: rio.Component,
    ) -> None:
        # Prepare
        layout = self._layouts_should[component._id_]
        layout.natural_height = 0

        # Pass on all space
        for child in iter_direct_tree_children(component):
            child_layout = self._layouts_should[child._id_]
            layout.natural_height = max(
                layout.natural_height,
                child_layout.requested_outer_height,
            )

    @specialized
    def _update_allocated_height(
        self,
        component: rio.Component,
    ) -> None:
        """
        Updates the allocated height of all of the component's children. This
        assumes that the component itself already has its requested height set.

        Furthermore, this also assigns the `left_in_viewport` attribute of all
        children.
        """
        # Default implementation: Trust the client
        for child in iter_direct_tree_children(component):
            child_layout_should = self._layouts_should[child._id_]
            child_layout_is = self._layouts_are[child._id_]

            child_layout_should.allocated_outer_height = (
                child_layout_is.allocated_outer_height
            )

            child_layout_should.top_in_viewport_outer = (
                child_layout_is.top_in_viewport_outer
            )

    def _update_allocated_height_HighLevelRootComponent(
        self,
        component: rio.components.root_components.HighLevelRootComponent,
    ) -> None:
        # Since the HighLevelRootComponent doesn't have a parent, it has to set
        # its own allocation
        layout_should = self._layouts_should[component._id_]
        layout_is = self._layouts_are[component._id_]

        # Because scrolling differs between debug mode and release mode (user
        # content scrolls vs browser scrolls), we'll just copy the values from
        # the client.
        layout_should.top_in_viewport_outer = layout_is.top_in_viewport_outer
        layout_should.top_in_viewport_inner = layout_is.top_in_viewport_inner

        layout_should.allocated_outer_height = layout_is.allocated_outer_height
        layout_should.allocated_inner_height = layout_is.allocated_inner_height

        # Then behave like a regular SingleContainer
        self._update_allocated_height_SingleContainer(component)

    def _update_allocated_height_Row(
        self,
        component: rio.Row,
    ) -> None:
        layout = self._layouts_should[component._id_]

        for child in component._iter_referenced_components_():
            child_layout = self._layouts_should[child._id_]
            child_layout.top_in_viewport_outer = layout.top_in_viewport_inner
            child_layout.allocated_outer_height = layout.allocated_inner_height

    def _update_allocated_height_Column(
        self,
        component: rio.Column,
    ) -> None:
        # Prepare
        layout = self._layouts_should[component._id_]

        child_heights: list[float] = []

        for child in component._iter_referenced_components_():
            child_layout = self._layouts_should[child._id_]
            child_heights.append(child_layout.requested_outer_height)

        # Update
        starts_and_sizes = _linear_container_get_major_axis_allocated_sizes(
            container_allocated_size=layout.allocated_inner_height,
            child_requested_sizes=child_heights,
            child_growers=[
                child.grow_y
                for child in component._iter_referenced_components_()
            ],
            spacing=component.spacing,
            proportions=component.proportions,
        )

        for child, (top, height) in zip(
            component._iter_referenced_components_(), starts_and_sizes
        ):
            child_layout = self._layouts_should[child._id_]
            child_layout.top_in_viewport_outer = (
                layout.top_in_viewport_inner + top
            )
            child_layout.allocated_outer_height = height

    def _update_allocated_height_Overlay(
        self,
        component: rio.Overlay,
    ) -> None:
        child_layout = self._layouts_should[component.content._id_]
        child_layout.top_in_viewport_outer = 0
        child_layout.allocated_outer_height = self.window_height

    def _update_allocated_height_SingleContainer(
        self,
        component: rio.Component,
    ) -> None:
        # Prepare
        layout = self._layouts_should[component._id_]

        # Pass on all space
        for child in iter_direct_tree_children(component):
            child_layout = self._layouts_should[child._id_]
            child_layout.top_in_viewport_outer = layout.top_in_viewport_inner
            child_layout.allocated_outer_height = layout.allocated_inner_height

    def debug_dump_json(
        self,
        which: t.Literal["should", "are"],
        out: t.IO[str],
    ) -> None:
        """
        Dumps the layouts to a JSON file.
        """
        # Export the layouts to a JSON file
        layouts = (
            self._layouts_should if which == "should" else self._layouts_are
        )

        # Convert the class instances to JSON
        result: list[dict[str, t.Any]] = []

        def dump_recursive(component: rio.Component) -> None:
            # Honor the filter function
            if not self._filter(component):
                return

            # Build the subresult
            layout = layouts[component._id_]
            value_json = {
                "id": component._id_,
                "type": type(component).__name__,
                **serialization.json_serde.as_json(layout),  # type: ignore (definitely a dict)
            }

            # Round floats
            for key, value in value_json.items():
                if isinstance(value, float):
                    value_json[key] = round(value, 1)

            # Store the subresult
            result.append(value_json)

            # Chain to children
            for child in iter_direct_tree_children(component):
                dump_recursive(child)

        dump_recursive(self.session._high_level_root_component)

        # Write the result
        json.dump(
            result,
            out,
            indent=4,
        )

    def debug_draw(
        self,
        which: t.Literal["should", "are"],
        *,
        pixels_per_unit: float = 10,
    ) -> PIL.Image.Image:
        """
        Draws the layout of all components to a raster image.
        """

        # Set up
        image = PIL.Image.new(
            "RGB",
            (
                round(self.window_width * pixels_per_unit),
                round(self.window_height * pixels_per_unit),
            ),
            color="white",  # type: ignore
        )

        draw = PIL.ImageDraw.Draw(image)

        layouts = (
            self._layouts_should if which == "should" else self._layouts_are
        )

        # How deep is the deepest nesting?
        def get_nesting(component: rio.Component, level: int) -> int:
            result = level

            for child in iter_direct_tree_children(component):
                result = max(result, get_nesting(child, level + 1))

            return result

        n_layers = get_nesting(self.session._high_level_root_component, 1)

        # Draw all components recursively
        def draw_component(
            component: rio.Component,
            level: int,
        ) -> None:
            # Honor the filter function
            if not self._filter(component):
                return

            # Determine the color. The deeper the darker
            nesting_fraction = level / (n_layers + 1)
            color_8bit = round(255 * nesting_fraction)
            color = (color_8bit, color_8bit, color_8bit)

            # Draw the component
            layout = layouts[component._id_]

            rect_left = layout.left_in_viewport_inner * pixels_per_unit
            rect_top = layout.top_in_viewport_inner * pixels_per_unit
            rect_right = (
                layout.left_in_viewport_inner + layout.allocated_inner_width
            ) * pixels_per_unit
            rect_bottom = (
                layout.top_in_viewport_inner + layout.allocated_inner_height
            ) * pixels_per_unit

            rect_right = max(rect_right, rect_left)
            rect_bottom = max(rect_bottom, rect_top)

            draw.rectangle(
                (
                    rect_left,
                    rect_top,
                    rect_right,
                    rect_bottom,
                ),
                fill=color,
            )

            # Label it
            label_str = type(component).__name__
            font = PIL.ImageFont.load_default()

            text_width = draw.textlength(label_str, font=font)
            text_height = pixels_per_unit * 1.5

            draw.text(
                (
                    rect_left + (rect_right - rect_left - text_width) / 2,
                    rect_top + (rect_bottom - rect_top - text_height) / 2,
                ),
                label_str,
                fill="blue",
                font=font,
            )

            # Chain to children
            for child in iter_direct_tree_children(component):
                draw_component(child, level + 1)

        draw_component(self.session._high_level_root_component, 1)

        # Done
        return image

    def print_tree(self) -> None:
        out = sys.stdout

        def print_worker(component: rio.Component, indent: str) -> None:
            # Honor the filter function
            if not self._filter(component):
                out.write("<filtered>")
                out.write("\n")
                return

            # Print the component
            out.write(type(component).__name__)
            out.write("\n")

            # Chain to children
            children = list(iter_direct_tree_children(component))
            for ii, child in enumerate(children):
                if ii == len(children) - 1:
                    out.write(indent + " └─ ")
                    child_indent = "    "
                else:
                    out.write(indent + " ├─ ")
                    child_indent = " │  "

                print_worker(child, indent + child_indent)

        print_worker(self.session._high_level_root_component, "")



import rio.testing


async def test_one_page_view() -> None:
    def build():
        return rio.Column(
            rio.Text("Welcome", style="heading1"),
            rio.PageView(),
        )

    app = rio.App(
        build=build,
        pages=[
            rio.ComponentPage(
                name="Home",
                url_segment="",
                build=rio.Spacer,
            ),
        ],
    )

    async with rio.testing.TestClient(app) as test_client:
        # Make sure the Spacer (which is located on the home page) exists
        test_client.get_component(rio.Spacer)


async def test_nested_page_views() -> None:
    def build():
        return rio.Column(
            rio.Text("Welcome", style="heading1"),
            rio.PageView(),
        )

    app = rio.App(
        build=build,
        pages=[
            rio.ComponentPage(
                name="Home",
                url_segment="",
                build=rio.PageView,
                children=[
                    rio.ComponentPage(
                        name="Inner Home",
                        url_segment="",
                        build=rio.Spacer,
                    ),
                ],
            ),
        ],
    )

    async with rio.testing.TestClient(app) as test_client:
        # Make sure the Spacer (which is located on the innermost page) exists
        test_client.get_component(rio.Spacer)



from __future__ import annotations

# <additional-imports>
import pandas as pd
import plotly.express as px

import rio

from .. import constants, data_models

# </additional-imports>


# <component>
class CryptoChart(rio.Component):
    """
    The CryptoChart class is a component of a dashboard application, designed to
    handle and display cryptocurrency-related data.


    ## Attributes

    `data`: Historical data of the fetched crypto coins.

    `coin`: Information about the selected coin.

    `horizon`: The time horizon for the chart.
    """

    data: pd.DataFrame
    coin: data_models.Coin = constants.MY_PORTFOLIO[0]  # Bitcoin as default

    horizon: int = 7

    def on_change_coin(self, ev: rio.DropdownChangeEvent) -> None:
        """
        Updates the coin, color and logo_url attributes based on
        the selected coin.

        ## Parameters

        `ev`: The event object containing the selected coin value.
        """
        # Find the coin object from MY_PORTFOLIO where name matches selected
        # value
        self.coin = next(
            coin for coin in constants.MY_PORTFOLIO if coin.name == ev.value
        )

    def build(self) -> rio.Component:
        """
        Creates a line plot of the last 50 data points of the selected coin.

        See the approx. layout below:
        ```
        ╔═════════════════════════════ CARD ══════════════════════════════╗
        ║ ┏━━━━━━━━━━━━━━━━━━━━━━━━━━ Column ━━━━━━━━━━━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ ┏━━━━━━━━━━━━━━━━━━━━━━━━━ Row ━━━━━━━━━━━━━━━━━━━━━━━━━━━┓ ┃ ║
        ║ ┃ ┃ ┏━━━━ Image ━━━━┓  ┏━━━━ Text ━━━━┓  ┏━━━ Dropdown ━━━┓ ┃ ┃ ║
        ║ ┃ ┃ ┃ url           ┃  ┃ str          ┃  ┃ dict           ┃ ┃ ┃ ║
        ║ ┃ ┃ ┗━━━━━━━━━━━━━━━┛  ┗━━━━━━━━━━━━━━┛  ┗━━━━━━━━━━━━━━━━┛ ┃ ┃ ║
        ║ ┃ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ┃ ║
        ║ ┃                                                             ┃ ║
        ║ ┃ ┏━━━━━━━━━━━━━━━━━━━━━━━━━ Plot ━━━━━━━━━━━━━━━━━━━━━━━━━━┓ ┃ ║
        ║ ┃ ┃                                                         ┃ ┃ ║
        ║ ┃ ┃ Plotly Express Figure                                   ┃ ┃ ║
        ║ ┃ ┃                                                         ┃ ┃ ║
        ║ ┃ ┃                                                         ┃ ┃ ║
        ║ ┃ ┃                                                         ┃ ┃ ║
        ║ ┃ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ╚═════════════════════════════════════════════════════════════════╝
        ```
        """
        # Get the device type from the session
        device = self.session[data_models.PageLayout].device

        # Set hovertemplate based on the time horizon
        if self.horizon == 1:
            hovertemplate = '<span style="font-size: 20px">$%{y:,.2f}</span><br>%{x|%b %-d, %H:%M}<extra></extra>'

        else:
            hovertemplate = '<span style="font-size: 20px">$%{y:,.2f}</span><br>%{x|%b %-d}<extra></extra>'

        # Create a line plot of the selected data points of the selected coin
        fig = px.line(
            self.data[self.coin.name].iloc[-self.horizon * 8 :],
            color_discrete_sequence=[self.coin.color],
        )

        # Make line thicker
        fig.update_traces(
            line=dict(width=3),
            hovertemplate=hovertemplate,
        )

        # Set x-axis labels to horizontal and remove labels
        fig.update_xaxes(
            tickangle=0,
            title_text="",
            tickfont=dict(color="white"),
        )
        fig.update_yaxes(
            title_text="",
            tickfont=dict(color="white"),
        )

        # strip down the rest of the plot
        fig.update_layout(
            showlegend=False,
            margin=dict(t=0, l=0, b=0, r=0),
            xaxis=dict(showgrid=False),
            yaxis=dict(showgrid=False),
        )

        # Create Desktop view
        if device == "desktop":
            return rio.Card(
                rio.Column(
                    rio.Row(
                        rio.Rectangle(
                            content=rio.Image(
                                rio.URL(
                                    self.coin.logo,
                                ),  # logo_url = e.g. "https://cryptologos.cc/logos/bitcoin-btc-logo.svg?v=029"
                                min_height=2,
                                min_width=2,
                                align_y=0.5,
                                align_x=0,
                                margin=0.5,
                            ),
                            fill=rio.Color.from_hex(self.coin.color).replace(
                                opacity=0.3
                            ),
                            corner_radius=self.session.theme.corner_radius_small,
                            align_x=0,
                            align_y=0.5,
                        ),
                        rio.Text(
                            self.coin.name.capitalize(),
                            style=rio.TextStyle(
                                font_size=1.2, font_weight="bold"
                            ),
                            align_x=0,
                            min_width=7,
                        ),
                        rio.Dropdown(
                            options={
                                coin.ticker: coin.name
                                for coin in constants.MY_PORTFOLIO
                            },
                            on_change=self.on_change_coin,
                            align_y=0.5,
                        ),
                        # Add a spacer to fill up the remaining space
                        rio.Spacer(),
                        # Create a switcher bar to change the time horizon, if
                        # the user selects a different time horizon, the chart
                        # will automatically update
                        rio.SwitcherBar(
                            values=[1, 7, 30, 90],
                            names=["1D", "7D", "1M", "3M"],
                            selected_value=self.bind().horizon,
                        ),
                        spacing=1,
                    ),
                    rio.Plot(
                        figure=fig,
                        corner_radius=0,
                        min_height=20,
                        min_width=10,
                        background=self.session.theme.neutral_color,
                    ),
                    spacing=1,
                    align_y=0.5,
                    margin=2,
                ),
            )

        # Build Mobile view
        # Same game, slightly different layout, and smaller plot hight
        return rio.Card(
            rio.Column(
                rio.Row(
                    rio.Rectangle(
                        content=rio.Image(
                            rio.URL(
                                self.coin.logo,
                            ),  # logo_url = e.g. "https://cryptologos.cc/logos/bitcoin-btc-logo.svg?v=029"
                            min_height=2,
                            min_width=2,
                            align_y=0.5,
                            align_x=0,
                            margin=0.5,
                        ),
                        fill=rio.Color.from_hex(self.coin.color).replace(
                            opacity=0.3
                        ),
                        corner_radius=self.session.theme.corner_radius_small,
                        align_x=0,
                        align_y=0.5,
                    ),
                    rio.Text(
                        self.coin.name.capitalize(),
                        style=rio.TextStyle(font_size=1.2, font_weight="bold"),
                        grow_x=True,
                        overflow="ellipsize",
                    ),
                    rio.Dropdown(
                        options={
                            coin.ticker: coin.name
                            for coin in constants.MY_PORTFOLIO
                        },
                        on_change=self.on_change_coin,
                        align_y=0.5,
                    ),
                    spacing=1,
                ),
                rio.SwitcherBar(
                    values=[1, 7, 30, 90],
                    names=["1D", "7D", "1M", "3M"],
                    selected_value=self.bind().horizon,
                ),
                rio.Plot(
                    figure=fig,
                    corner_radius=0,
                    min_height=15,
                    background=self.session.theme.neutral_color,
                ),
                spacing=1,
                align_y=0.5,
                margin=2,
            ),
        )


# </component>



import abc
import typing as t

import rio

__all__ = [
    "Extension",
]


class Extension(abc.ABC):
    """
    Base class for all Rio extensions.

    Rio extensions/plugins are classes that extend the functionality of Rio
    beyond what the core library provides. Extensions are Python classes, that
    inherit from `rio.Extension`, and are registered with the Rio app before it
    is run. They allow you to e.g. add additional pages, log system resources as
    new users connect/disconnect, and more.

    Methods in extensions can be decorated with one of the events defined in
    `rio.extension_event.*`, such as `rio.extension_event.on_session_start`.
    Those functions will then be called whenever the event occurs - in this
    case, when a user visits the app.

    ## Examples

    This minimal extension will log a message whenever a user visits or leaves
    the app:

    ```py
    # This can be any component. This one just displays a text
    class MyRoot(rio.Component):
        def build(self) -> rio.Component:
            return rio.Text(
                "Hello, World!",
                align_x=0.5,
                align_y=0.5,
            )

    # The extension itself. It has an event handler for when a session starts
    # and one for when a session ends. They both print a simple message to the
    # console when they are called. You can modify these to do anything you'd
    # like - connect to a database, change the app's default attachments,
    # whatever.
    class MyExtension(rio.Extension):
        @rio.extension_event.on_session_start
        def on_session_start(
            self,
            event: rio.ExtensionSessionStartEvent,
        ) -> None:
            print("Extension: Session started")

        @rio.extension_event.on_session_close
        def on_session_close(
            self,
            event: rio.ExtensionSessionCloseEvent,
        ) -> None:
            print("Extension: Session closed")


    # Instantiate the app and add the extension to it
    app = rio.App(
        build=MyRoot,
        extensions=[
            MyExtension(),
        ],
    )
    ```

    ## Metadata

    `public`: False
    """

    _rio_on_app_start_event_handlers_: t.ClassVar[
        list[t.Callable[[rio.extension_event.ExtensionAppStartEvent], None]]
    ]

    _rio_on_app_close_event_handlers_: t.ClassVar[
        list[t.Callable[[rio.extension_event.ExtensionAppCloseEvent], None]]
    ]

    _rio_on_session_start_event_handlers_: t.ClassVar[
        list[t.Callable[[rio.extension_event.ExtensionSessionStartEvent], None]]
    ]

    _rio_on_session_close_event_handlers_: t.ClassVar[
        list[t.Callable[[rio.extension_event.ExtensionSessionCloseEvent], None]]
    ]

    def __init__(self) -> None:
        pass

    @property
    def name(self) -> str:
        """
        A human-readable name for the extension.

        Returns the name of the extension. This is sometimes used when referring
        to the plugin, for example in an admin interface that lists all active
        plugins in an app.

        Override this to change your extension's name.
        """
        # By default, derive one
        name = self.__class__.__name__
        name = name.removesuffix("Extension")
        return name

    @property
    def description(self) -> str:
        """
        A detailed description of the extension.

        Returns a description of the extension. This is sometimes used when
        referring to the plugin, for example in an admin interface that lists
        all active plugins in an app.

        Override this to change your extension's description.
        """
        return ""

    @property
    def icon(self) -> str:
        """
        A visual icon for the extension.

        Returns the icon of the extension, in standard Rio notation, for example
        `"material/extension"`. This is sometimes used when referring to the
        plugin, for example in an admin interface that lists all active plugins
        in an app.

        Override this to change your extension's icon.
        """
        # TODO: Shouldn't this return a path, asset or similar? Seems unlikely
        # that plugins just want to use material icons.

        return "material/extension"



import asyncio
import json
import typing as t

import typing_extensions as te
from uniserde import JsonDoc

from . import abstract_transport

__all__ = ["MessageRecorderTransport"]


class MessageRecorderTransport(abstract_transport.AbstractTransport):
    def __init__(
        self, *, process_sent_message: t.Callable[[JsonDoc], None] | None = None
    ) -> None:
        super().__init__()

        self.process_sent_message = process_sent_message

        self.sent_messages = list[JsonDoc]()
        self._responses = asyncio.Queue[
            JsonDoc
            | type[abstract_transport.TransportInterrupted]
            | type[abstract_transport.TransportClosedIntentionally]
        ]()

    @te.override
    async def send(self, msg: str) -> None:
        parsed_msg = json.loads(msg)
        self.sent_messages.append(parsed_msg)

        if self.process_sent_message is not None:
            self.process_sent_message(parsed_msg)

    @te.override
    async def receive(self) -> str:
        response = await self._responses.get()

        if isinstance(response, type) and issubclass(response, Exception):
            raise response()

        return json.dumps(response)

    def close(self) -> None:
        self._responses.put_nowait(
            abstract_transport.TransportClosedIntentionally
        )
        self.closed.set()

    def queue_response(
        self,
        response: JsonDoc
        | type[abstract_transport.TransportInterrupted]
        | type[abstract_transport.TransportClosedIntentionally],
    ) -> None:
        self._responses.put_nowait(response)



from __future__ import annotations

# <additional-imports>
import rio

from .. import components as comps
from .. import data_models, persistence

# </additional-imports>


# <component>
def guard(event: rio.GuardEvent) -> str | None:
    """
    A guard which only allows the user to access this page if they are not
    logged in yet. If the user is already logged in, the login page will be
    skipped and the user will be redirected to the home page instead.
    """
    # Check if the user is authenticated by looking for a user session
    try:
        event.session[data_models.AppUser]

    # User is not logged in, no redirection needed
    except KeyError:
        return None

    # User is logged in, redirect to the home page
    return "/app/home"


@rio.page(
    name="Login",
    url_segment="",
    guard=guard,
)
class LoginPage(rio.Component):
    """
    Login page for accessing the website.
    """

    # These are used to store the currently entered values from the user
    username: str = ""
    password: str = ""

    error_message: str = ""
    popup_open: bool = False

    _currently_logging_in: bool = False

    async def login(self, _: rio.TextInputConfirmEvent | None = None) -> None:
        """
        Handles the login process when the user submits their credentials.

        It will check if the user exists and if the password is correct. If the
        user exists and the password is correct, the user will be logged in and
        redirected to the home page.
        """
        try:
            # Inform the user that something is happening
            self._currently_logging_in = True
            self.force_refresh()

            #  Try to find a user with this name
            pers = self.session[persistence.Persistence]

            try:
                user_info = await pers.get_user_by_username(
                    username=self.username
                )
            except KeyError:
                self.error_message = "Invalid username. Please try again or create a new account."
                return

            # Make sure their password matches
            if not user_info.password_equals(self.password):
                self.error_message = "Invalid password. Please try again or create a new account."
                return

            # The login was successful
            self.error_message = ""

            # Create and store a session
            user_session = await pers.create_session(
                user_id=user_info.id,
            )

            # Attach the session and userinfo. This indicates to any other
            # component in the app that somebody is logged in, and who that is.
            self.session.attach(user_session)
            self.session.attach(user_info)

            # Permanently store the session token with the connected client.
            # This way they can be recognized again should they reconnect later.
            settings = self.session[data_models.UserSettings]
            settings.auth_token = user_session.id
            self.session.attach(settings)

            # The user is logged in - no reason to stay here
            self.session.navigate_to("/app/home")

        # Done
        finally:
            self._currently_logging_in = False

    def on_open_popup(self) -> None:
        """
        Opens the sign-up popup when the user clicks the sign-up button
        """
        self.popup_open = True

    def build(self) -> rio.Component:
        # Determine the layout based on the window width
        desktop_layout = self.session.window_width > 30

        return rio.Card(
            rio.Column(
                rio.Text("Sign In", style="heading1", justify="center"),
                # Show error message if there is one
                #
                # Banners automatically appear invisible if they don't have
                # anything to show, so there is no need for a check here.
                rio.Banner(
                    text=self.error_message,
                    style="danger",
                    margin_top=1,
                ),
                # Create the login form consisting of a username and password
                # input field, a login button and a sign up button
                rio.TextInput(
                    text=self.bind().username,
                    label="Username",
                    # ensure the login function is called when the user presses enter
                    on_confirm=self.login,
                ),
                rio.TextInput(
                    text=self.bind().password,
                    label="Password",
                    # Mark the password field as secret so the password is
                    # hidden while typing
                    is_secret=True,
                    # Ensure the login function is called when the user presses
                    # enter
                    on_confirm=self.login,
                ),
                rio.Button(
                    "Sign In",
                    on_press=self.login,
                    is_loading=self._currently_logging_in,
                ),
                # Create a sign up button that opens a pop up with a sign up
                # form
                rio.Popup(
                    anchor=rio.Button(
                        "Create an Account",
                        on_press=self.on_open_popup,
                        style="minor",
                    ),
                    content=comps.UserSignUpForm(
                        # Bind `popup_open` to the `popup_open` attribute of
                        # the login page. This way the page's attribute will
                        # always have the same value as that of the form,
                        # even when one changes.
                        popup_open=self.bind().popup_open,
                    ),
                    position="fullscreen",
                    is_open=self.popup_open,
                    color="none",
                ),
                spacing=1,
                margin=2 if desktop_layout else 1,
            ),
            margin_x=0.5,
            align_y=0.5,
            align_x=0.5 if desktop_layout else None,
            min_width=24 if desktop_layout else 0,
        )


# </component>



import rio

# <additional-imports>
from .. import data_models

# </additional-imports>


# <component>
class StyledTransaction(rio.Component):
    """
    A component that visually represents a transaction in a styled format.

    This component displays details such as:
    - Transaction type (Buy/Sell) with corresponding icon.
    - The asset name and its ticker.
    - The amount transacted and the transaction date.


    ## Attributes:

    - `transaction`: The transaction data to be displayed.
    """

    transaction: data_models.Transaction

    def build(self) -> rio.Component:
        # Determine transaction type and set corresponding text and icon
        # Sell transaction
        if self.transaction.amount < 0:
            text = "Sell"
            icon = rio.Icon(
                "material/arrow_downward", min_height=2, min_width=2, margin=0.5
            )

        # Buy transaction
        else:
            text = "Buy"
            icon = rio.Icon(
                "material/arrow_upward", min_height=2, min_width=2, margin=0.5
            )

        # Construct the transaction row layout
        return rio.Row(
            # Left rectangle with icon and background
            rio.Rectangle(
                content=icon,
                fill=self.session.theme.primary_color.replace(opacity=0.1),
                corner_radius=9999,
                align_y=0.5,
            ),
            # Column with transaction text details
            rio.Column(
                rio.Text(
                    f"{text} {self.transaction.name}",
                    font_weight="bold",
                    overflow="ellipsize",
                ),
                rio.Text(
                    f"{self.transaction.ticker} / USD",
                    style="dim",
                    font_size=0.8,
                ),
                grow_x=True,
            ),
            # Column with amount and date details
            rio.Column(
                rio.Text(
                    f"{self.transaction.amount:.5f}",
                    font_weight="bold",
                    justify="right",
                ),
                rio.Text(
                    self.transaction.date.strftime("%b %#d, %Y %H:%M"),
                    style="dim",
                    font_size=0.8,
                    justify="right",
                ),
            ),
            spacing=1,
        )


# </component>



from pathlib import Path

from . import data_models

PROJECT_ROOT_DIR = Path(__file__).resolve().parent
ASSETS_DIR = PROJECT_ROOT_DIR / "assets"


SERVICE_POWER = [
    (
        "material/child_care",
        "Easy to use",
        "Lorem ipsum dolor sit amet, consetetur sadipscing elitr.",
    ),
    (
        "material/check",
        "Reliable",
        "At vero eos et accusam et justo duo dolores et ea rebum.",
    ),
    (
        "material/safety_check",
        "Secure",
        "Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam voluptua.",
    ),
]


SERVICE_SPEED = [
    (
        "material/rocket_launch",
        "Fast",
        "Lorem ipsum dolor sit amet, consetetur sadipscing elitr.",
    ),
    (
        "material/savings",
        "Affordable",
        "Sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam erat.",
    ),
    (
        "material/signal_cellular_alt",
        "Scalable",
        "At vero eos et accusam et justo duo dolores et ea rebum.",
    ),
]


TESTIMONIALS = [
    data_models.Testimonial(
        quote='"Duis autem vel eum iriure dolor in hendrerit in vulputate velit esse molestie consequat, vel illum dolore eu feugiat nulla facilisis at vero eros et accumsan et iusto odio dignissim qui blandit praesent luptatum zzril delenit augue duis dolore te feugait nulla facilisi. Lorem ipsum dolor sit amet, consectetuer adipiscing elit, sed diam nonummy nibh euismod tincidunt ut laoreet dolore magna aliquam erat volutpat."',
        name="Sophia Bennett",
        company="Nexora Inc.",
        image="testimonial_1.png",
    ),
    data_models.Testimonial(
        quote='"Duis autem vel eum iriure dolor in hendrerit in vulputate velit esse molestie consequat, vel illum dolore eu feugiat nulla facilisis."',
        name="Lorenzo Wagner",
        company="SYNC Inc.",
        image="testimonial_2.png",
    ),
    data_models.Testimonial(
        quote='"Ut wisi enim ad minim veniam, quis nostrud exerci tation ullamcorper suscipit lobortis nisl ut aliquip ex ea commodo consequat. Duis autem vel eum iriure dolor in hendrerit in vulputate velit esse molestie consequat, vel illum dolore eu feugiat nulla facilisis at vero eros et accumsan et iusto odio dignissim qui blandit praesent luptatum zzril delenit augue duis dolore te feugait nulla facilisi."',
        name="Emily Hart",
        company="ACME Inc.",
        image="testimonial_3.png",
    ),
    data_models.Testimonial(
        quote='"Duis autem vel eum iriure dolor in hendrerit in vulputate velit esse molestie consequat, vel illum dolore eu feugiat nulla facilisis at vero eros et accumsan et iusto odio dignissim qui blandit praesent luptatum zzril delenit augue duis dolore te feugait nulla facilisi. Lorem ipsum dolor sit amet"',
        name="Oliver Grant",
        company="Vistral Inc.",
        image="testimonial_4.png",
    ),
    data_models.Testimonial(
        quote='"Ut wisi enim ad minim veniam, quis nostrud exerci tation ullamcorper suscipit lobortis nisl ut aliquip ex ea commodo consequat. Duis autem vel eum iriure dolor in."',
        name="Liam Carter",
        company="Trivex Inc.",
        image="testimonial_5.png",
    ),
    data_models.Testimonial(
        quote='"Ut wisi enim ad minim veniam, quis nostrud exerci tation ullamcorper suscipit lobortis nisl ut aliquip ex ea commodo consequat. Duis autem vel eum iriure dolor in hendrerit in vulputate velit esse molestie consequat, vel illum dolore eu feugiat nulla facilisis at vero eros et accumsan et iusto odio dignissim qui blandit praesent luptatum zzril delenit augue duis dolore te feugait nulla facilisi."',
        name="James Turner",
        company="Clyro Inc.",
        image="testimonial_6.png",
    ),
]


BLOG_POSTS = [
    # 1. Matterhorn
    data_models.BlogPost(
        title="Adventure in the Swiss Alps",
        description="Exploring the Summit of the Matterhorn",
        image="https://fastly.picsum.photos/id/866/1200/600.jpg?hmac=gYnC9Wd2DD8keY1moAbMLvF5p1QkEIEKoLH6Jr1jLLg",
        category="Nature",
        author="Mia Thompson",
        author_image="blog_author_1.png",
        date="Nov 29, 2024",
        url_segment="/adventure-in-the-swiss-alps",
    ),
    # 2. Stone Water
    data_models.BlogPost(
        title="Water and the healing power of stones",
        description="What does the sound of water and the energy of stones do to us?",
        image="https://fastly.picsum.photos/id/609/1200/600.jpg?hmac=s2I-AQvNCACy1MFdRCqEbR5AC6g2PGVft36B1GqkavI",
        category="Nature",
        author="Jack Harrison",
        author_image="blog_author_2.png",
        date="Sep 6, 2024",
        url_segment="/power-of-stones-and-water",
    ),
    # 3. Mountain Cliff
    data_models.BlogPost(
        title="Climate change becomes visible",
        description="Get to know the effects of climate change",
        image="https://fastly.picsum.photos/id/511/1200/600.jpg?hmac=671MXvebd-4LD-FNfrXCz88-NvLqfzLuDOpg1Zgfl2I",
        category="Climate Change",
        author="Henry Scott",
        author_image="blog_author_3.png",
        date="Jan 15, 2024",
        url_segment="/climate-change-effects",
    ),
    # 4. Beach
    data_models.BlogPost(
        title="The 10 most beautiful beaches",
        description="From Navagio Beach to the Maldives - the most beautiful beaches in the world",
        image="https://fastly.picsum.photos/id/640/1200/600.jpg?hmac=jjcV1z6Vi_dLyTdWLTS8YNiQ6MNzPoggXvEF0Lji7dE",
        category="Travelling",
        author="Grace Robinson",
        author_image="blog_author_4.png",
        date="Dez 1, 2023",
        url_segment="/most-beautiful-beaches",
    ),
    # 5. Camera
    data_models.BlogPost(
        title="Unveiling the Secrets of Photography",
        description="The Journey to Becoming a Professional Photographer",
        image="https://fastly.picsum.photos/id/435/1200/600.jpg?hmac=GMMLcKhgXuGoJqfcQDZvjL49DfBVo2ci8vV6gSdfnBM",
        category="Photography",
        author="Ava Mitchell",
        author_image="blog_author_5.png",
        date="Jun 20, 2023",
        url_segment="/secrets-of-photography",
    ),
    # 6. Table
    data_models.BlogPost(
        title="Workacholic or Workahappy?",
        description="Embarking on a Journey to a Healthier Work Life",
        image="https://fastly.picsum.photos/id/7/1200/600.jpg?hmac=9BZXw2Tg--BqxsEs_QtlArAbdjZ0i9kPoW9GM4O7LCs",
        category="Work Life Balance",
        author="Mason Reed",
        author_image="blog_author_6.png",
        date="May 10, 2023",
        url_segment="/work-life-balance",
    ),
    # 7. Sunset
    data_models.BlogPost(
        title="Discovering the Beauty of Sunsets",
        description="The energy of the sun and the beauty of the sky",
        image="https://fastly.picsum.photos/id/900/1200/600.jpg?hmac=zbCWU02ATL9Ga-tgU6dENO1sllOymdxjpRxVWf6aeVA",
        category="Travelling",
        author="Owen Brooks",
        author_image="blog_author_7.png",
        date="Feb 25, 2023",
        url_segment="/beauty-of-sunsets",
    ),
]

# Dictionary mapping blog post URL segments to their corresponding BlogPost
# objects. Only includes posts with non-empty url_segments. Access pattern:
# BLOG_POSTS_BY_URL["/url-segment"]
BLOG_POSTS_BY_URL = {
    post.url_segment: post for post in BLOG_POSTS if post.url_segment
}



from .navbar import Navbar as Navbar
from .news_article import NewsArticle as NewsArticle
from .root_component import RootComponent as RootComponent
from .testimonial import Testimonial as Testimonial
from .user_sign_up_form import UserSignUpForm as UserSignUpForm



from __future__ import annotations

# <additional-imports>
import rio

from ... import components as comps
from ... import data_models, utils

# </additional-imports>

# <component>
# List of image URLs to be used in the page
image_url: list[str] = [
    "https://images.pexels.com/photos/1264210/pexels-photo-1264210.jpeg?auto=compress&cs=tinysrgb&w=1260&h=750&dpr=1",
    "https://images.pexels.com/photos/243757/pexels-photo-243757.jpeg?auto=compress&cs=tinysrgb&w=1260&h=750&dpr=1",
]


@rio.page(
    name="Secrets of Photography",
    url_segment="secrets-of-photography",
)
class WaterAndHealingPowerOfStones(rio.Component):
    """
    WaterAndHealingPowerOfStones Page Component.

    Displays content about water and healing stones with a responsive layout
    that adapts between desktop and mobile views.
    """

    def build(self) -> rio.Component:
        # Retrieve the device type from the session data
        device = self.session[data_models.PageLayout].device

        # get the URL name of the active page
        url_name = self.session.active_page_url.name
        # Get single blog post by URL key
        blog_post = utils.BLOG_POSTS_BY_URL["/" + url_name]

        if device == "desktop":
            min_img_width = 35  # Fixed image width for desktop
            min_img_height = 20  # Fixed image height for desktop
            spacing = 2  # Spacing between images for desktop
        else:
            margin_x = 1
            # Calculate the minimum image width for mobile based on the window
            # width and margin
            min_img_width = self.session.window_width - margin_x * 2
            min_img_height = 13  # Fixed image height for mobile
            spacing = 0.5  # Spacing between images for mobile

        # Create the markdown content
        markdown_content = rio.Markdown(
            """
# Lorem ipsum dolor sit amet, consetetur sadipscing elitr

Duis autem vel eum iriure dolor in hendrerit in vulputate velit esse molestie
consequat, vel illum dolore eu feugiat nulla facilisis at vero eros et accumsan
et iusto odio dignissim qui blandit praesent luptatum zzril delenit augue duis
dolore te feugait nulla facilisi. Lorem ipsum dolor sit amet, consectetuer
adipiscing elit, sed diam nonummy nibh euismod tincidunt ut laoreet dolore magna
aliquam erat volutpat. 

Ut wisi enim ad minim veniam, quis nostrud exerci tation ullamcorper suscipit
lobortis nisl ut aliquip ex ea commodo consequat. Duis autem vel eum iriure
dolor in hendrerit in vulputate velit esse molestie consequat, vel illum dolore
eu feugiat nulla facilisis at vero eros et accumsan et iusto odio dignissim qui
blandit praesent luptatum zzril delenit augue duis dolore te feugait nulla
facilisi.
            """
        )

        # Create the image content
        image_content = rio.Image(
            rio.URL(image_url[0]),
            fill_mode="zoom",
            corner_radius=self.session.theme.corner_radius_medium,
            min_width=min_img_width,
            min_height=min_img_height,
        )

        markdown_content2 = rio.Markdown(
            """
# Lorem ipsum dolor sit amet, consetetur sadipscing elitr

Duis autem vel eum iriure dolor in hendrerit in vulputate velit esse molestie
consequat, vel illum dolore eu feugiat nulla facilisis at vero eros et accumsan
et iusto odio dignissim qui blandit praesent luptatum zzril delenit augue duis
dolore te feugait nulla facilisi. Lorem ipsum dolor sit amet, consectetuer
adipiscing elit, sed diam nonummy nibh euismod tincidunt ut laoreet dolore magna
aliquam erat volutpat. 

Ut wisi enim ad minim veniam, quis nostrud exerci tation ullamcorper suscipit
lobortis nisl ut aliquip ex ea commodo consequat. Duis autem vel eum iriure
dolor in hendrerit in vulputate velit esse molestie consequat, vel illum dolore
eu feugiat nulla facilisis at vero eros et accumsan et iusto odio dignissim qui
blandit praesent luptatum zzril delenit augue duis dolore te feugait nulla
facilisi.
            """
        )

        # Create the second image content
        image_content2 = rio.Image(
            rio.URL(image_url[1]),
            fill_mode="zoom",
            corner_radius=self.session.theme.corner_radius_medium,
            min_width=min_img_width,
            min_height=min_img_height,
        )

        # Create the main content based on device type
        if device == "desktop":
            content = rio.Column(
                comps.BlogHeader(blog_post=blog_post),
                rio.Row(
                    markdown_content,
                    image_content,
                    spacing=spacing,
                ),
                rio.Row(
                    image_content2,
                    markdown_content2,
                    spacing=spacing,
                ),
                spacing=4,  # Increased spacing between rows for better readability
            )
        else:
            content = rio.Column(
                comps.BlogHeader(blog_post=blog_post),
                markdown_content,
                image_content,
                markdown_content2,  # Fixed: Using markdown_content2 instead of duplicate
                image_content2,  # Fixed: Using image_content2 instead of duplicate
                spacing=spacing,
            )

        # Return the appropriate container based on device type
        if device == "desktop":
            return rio.Container(
                content=content,
                align_x=0.5,  # Center the container horizontally
                min_width=80,  # Fixed width for desktop layout
            )
        else:
            return rio.Container(
                content=content,
                margin_x=1,  # Horizontal margin for mobile layout
            )


# </component>



from __future__ import annotations

import dataclasses
import typing as t

import imy.docstrings

import rio

from . import component, number_input, switch

__all__ = [
    "FormBuilder",
]

T = t.TypeVar("T")


@t.final
@imy.docstrings.mark_constructor_as_private
@dataclasses.dataclass
class FormChangeEvent:
    """
    Holds information regarding a FormBuilder's change event.

    This is a simple dataclass that stores useful information for when the user
    changes a value in a `FormBuilder`. You'll typically receive this as
    argument in `on_change` events.

    ## Attributes

    `key`: They key of the component that has changed.

    `is_valid`: Whether all values in the form builder are now valid.

    `validation_error`: This is `None` if the form is valid, otherwise it
        contains a string with the error message.
    """

    key: int | str | None

    is_valid: bool

    validation_error: None | str


class FormBuilder(component.Component):
    """
    TODO

    ## Metadata

    `public`: False
    """

    _: dataclasses.KW_ONLY

    heading: str | None = None
    heading_style: t.Literal["heading1", "heading2", "heading3"] = "heading2"

    is_valid: bool = True
    validation_error: None | str = None

    on_change: rio.EventHandler[[FormChangeEvent]] = None

    # All fields in the form are added to this list. Each entry is a tuple
    # of
    #
    # - The component to display for the field
    # - The key of the field
    # - The validation function
    # - A function to call when the field changes
    #
    # The index in the field is used as unique identifier for the field.
    _all_fields: list[
        tuple[
            rio.Component,
            int | str | None,
            t.Callable[[t.Any], str | None] | None,
            rio.EventHandler[t.Any],
        ]
    ] = dataclasses.field(
        default_factory=list,
        init=False,
    )

    # Maps field indices to their validation errors, if any. If no errors
    # show up in here the form is valid.
    _all_errors: dict[int, str] = dataclasses.field(
        default_factory=dict,
        init=False,
    )

    async def _on_change(self, field_index: int, new_value: t.Any) -> None:
        """
        Call this whenever a value has changed. It will re-run the validator and
        trigger events as appropriate.
        """
        # Get information about this field
        _, key, validator, field_change_handler = self._all_fields[field_index]

        # Re-validate the form
        if validator is None:
            validation_error = None
        else:
            validation_error = validator(new_value)

        # Update the validation state
        if validation_error is None:
            self._all_errors.pop(field_index, None)

            if self._all_errors:
                self.validation_error = next(iter(self._all_errors.values()))
            else:
                self.validation_error = None
        else:
            self._all_errors[field_index] = validation_error
            self.validation_error = validation_error

        self.is_valid = self.validation_error is None

        # Trigger the field's change handler
        await self.call_event_handler(field_change_handler, new_value)

        # Trigger the form's change event
        await self.call_event_handler(
            self.on_change,
            FormChangeEvent(
                key=key,
                is_valid=self.is_valid,
                validation_error=self.validation_error,
            ),
        )

    def _add_custom_field(
        self,
        component: rio.Component,
        key: int | str | None,
        validate: t.Callable[[T], str | None] | None,
        on_change: rio.EventHandler[T],
    ) -> int:
        """
        Adds a field to the form builder. Returns the index of the field.
        """
        self._all_fields.append((component, key, validate, on_change))
        return len(self._all_fields) - 1

    def _add_labelled_field(
        self,
        label: str,
        value_component: rio.Component,
        key: int | str | None,
        validate: t.Callable[[T], str | None] | None,
        on_change: rio.EventHandler[T],
    ) -> int:
        """
        Adds a component in "standard style". This will have a text label added
        as well as the component itself.
        """
        # Create the component
        column = rio.Row(
            rio.Text(label),
            value_component,
            spacing=1,
        )

        # Add the component
        self._all_fields.append((column, key, validate, on_change))
        return len(self._all_fields) - 1

    def add_bool(
        self,
        label: str,
        value: bool,
        *,
        style: t.Literal["switch", "checkbox"] = "switch",
        key: int | str | None = None,
        validate: t.Callable[[bool], str | None] | None = None,
        on_change: rio.EventHandler[bool] = None,
    ) -> None:
        """
        Adds a boolean field to the form.

        This adds a field to the form that allows the user to toggle a boolean
        value. You can choose between a switch or a checkbox style.

        ## Parameters

        `label`: The label to display next to the field.

        `value`: The initial value of the field. Use `self.bind().value` to
            automatically receive changes from this field.

        `style`: How to display the field. You can choose between a "switch" or
            "checkbox" style.

        `key`: A unique key for this field. This is useful if you want to
            identify the field when the form changes.

        `validate`: A function that validates the value. This should return a
            string with an error message if the value is invalid, otherwise
            `None`.

        `on_change`: A function that is called whenever the value changes. This
            function receives the new value as argument.
        """

        # Create the component
        index = len(self._all_fields)

        if style == "switch":
            component = switch.Switch(
                is_on=value,
                on_change=lambda e: self._on_change(index, e.is_on),
            )
        elif style == "checkbox":
            component = rio.Checkbox(
                is_on=value,
                on_change=lambda e: self._on_change(index, e.is_on),
            )
        else:
            raise ValueError(f"Invalid style: {style}")

        # Add the component
        self._add_labelled_field(
            label,
            component,
            key,
            validate,
            on_change,
        )

    def add_number(
        self,
        label: str,
        value: int | float,
        *,
        decimals: int = 2,
        key: int | str | None = None,
        validate: t.Callable[[int | float], str | None] | None = None,
        on_change: rio.EventHandler[int | float] = None,
    ) -> None:
        # Create the component
        index = len(self._all_fields)

        component = number_input.NumberInput(
            value=value,
            decimals=decimals,
            on_change=lambda e: self._on_change(index, e.value),
        )

        # Add the component
        self._add_labelled_field(
            label,
            component,
            key,
            validate,
            on_change,
        )

    def build(self) -> rio.Component:
        column = rio.Column(
            spacing=1,
        )

        # Heading, if any
        if self.heading is not None:
            column.add(
                rio.Text(
                    self.heading,
                    style=self.heading_style,
                ),
            )

        # Add all fields to the column
        for component, *_ in self._all_fields:
            column.add(component)

        # Done
        return column



from __future__ import annotations

import dataclasses
import typing as t

import imy.docstrings
from uniserde import JsonDoc

import rio

from .keyboard_focusable_components import KeyboardFocusableFundamentalComponent

__all__ = [
    "MultiLineTextInput",
    "MultiLineTextInputChangeEvent",
    "MultiLineTextInputConfirmEvent",
]


@t.final
@imy.docstrings.mark_constructor_as_private
@dataclasses.dataclass
class MultiLineTextInputChangeEvent:
    """
    Holds information regarding a text input change event.

    This is a simple dataclass that stores useful information for when the user
    changes the text in a `MultiLineTextInput`. You'll typically receive this as
    argument in `on_change` events.

    ## Attributes

    `text`: The new `text` of the `MultiLineTextInput`.
    """

    text: str


@t.final
@imy.docstrings.mark_constructor_as_private
@dataclasses.dataclass
class MultiLineTextInputConfirmEvent:
    """
    Holds information regarding a text input confirm event.

    This is a simple dataclass that stores useful information for when the user
    confirms the text in a `MultiLineTextInput`. You'll typically receive this
    as argument in `on_confirm` events.

    ## Attributes

    `text`: The new `text` of the `MultiLineTextInput`.
    """

    text: str


@t.final
class MultiLineTextInput(KeyboardFocusableFundamentalComponent):
    """
    A user-editable text field.

    `MultiLineTextInput` is a text input field similar to the regular
    `TextInput`, but allows the user to enter multiple lines of text.

    ## Attributes

    `text`: The text currently entered by the user.

    `label`: A short text to display next to the text input.

    `accessibility_label`: A short text describing the text input for screen
        readers. If omitted, the `label` text is used.

    `style`: Changes the visual appearance of the text input.

    `is_sensitive`: Whether the text input should respond to user input.

    `is_valid`: Visually displays to the user whether the current text is
        valid. You can use this to signal to the user that their input needs
        to be changed.

    `auto_adjust_height`: Whether to automatically grow the text input to
        accommodate all the text written in it.

    `on_change`: Triggered when the user changes the text.

    `on_confirm`: Triggered when the user explicitly confirms their input,
        such as by pressing the "Enter" key. You can use this to trigger
        followup actions, such as logging in or submitting a form.


    ## Examples

    Here's a simple example that allows the user to enter a value and displays
    it back to them:

    ```python
    class MyComponent(rio.Component):
        text: str = "Hello, World!"

        def build(self) -> rio.Component:
            return rio.Column(
                rio.MultiLineTextInput(
                    # In order to retrieve a value from the component, we'll
                    # use an attribute binding. This way our own value will
                    # be updated whenever the user changes the text.
                    text=self.bind().text,
                    label="Enter a Text",
                ),
                rio.Text(f"You've typed: {self.text}"),
            )
    ```

    Alternatively you can also attach an event handler to react to changes. This
    is a little more verbose, but allows you to run arbitrary code when the user
    changes the text:

    ```python
    class MyComponent(rio.Component):
        text: str = "Hello, World!"

        def on_value_change(self, event: rio.MultiLineTextInputChangeEvent):
            # This function will be called whenever the input's value
            # changes. We'll display the new value in addition to updating
            # our own attribute.
            self.text = event.text
            print(f"You've typed: {self.text}")

        def build(self) -> rio.Component:
            return rio.MultiLineTextInput(
                text=self.text,
                label="Enter a Text",
                on_change=self.on_value_change,
            )
    ```
    """

    text: str = ""
    _: dataclasses.KW_ONLY
    label: str = ""
    accessibility_label: str = ""
    is_sensitive: bool = True
    is_valid: bool = True
    auto_adjust_height: bool = True

    on_change: rio.EventHandler[MultiLineTextInputChangeEvent] = None
    on_confirm: rio.EventHandler[MultiLineTextInputConfirmEvent] = None

    # Note the lack of the `"pill"` style. It looks silly with tall components
    # so is intentionally omitted here.
    style: t.Literal["underlined", "rounded"] = "underlined"

    def _validate_delta_state_from_frontend(self, delta_state: JsonDoc) -> None:
        if not set(delta_state) <= {"text"}:
            raise AssertionError(
                f"Frontend tried to change `{type(self).__name__}` state: {delta_state}"
            )

        if "text" in delta_state and not self.is_sensitive:
            raise AssertionError(
                f"Frontend tried to set `MultiLineTextInput.text` even though `is_sensitive` is `False`"
            )

    async def _call_event_handlers_for_delta_state(
        self, delta_state: JsonDoc
    ) -> None:
        # Trigger on_change event
        try:
            new_value = delta_state["text"]
        except KeyError:
            pass
        else:
            assert isinstance(new_value, str), new_value
            await self.call_event_handler(
                self.on_change,
                MultiLineTextInputChangeEvent(new_value),
            )

        self._apply_delta_state_from_frontend(delta_state)

    async def _on_message_(self, msg: t.Any) -> None:
        # Listen for messages indicating the user has confirmed their input
        #
        # In addition to notifying the backend, these also include the input's
        # current value. This ensures any event handlers actually use the up-to
        # date value.
        assert isinstance(msg, dict), msg

        self._apply_delta_state_from_frontend({"text": msg["text"]})

        await self.call_event_handler(
            self.on_change,
            MultiLineTextInputChangeEvent(self.text),
        )

        await self.call_event_handler(
            self.on_confirm,
            MultiLineTextInputConfirmEvent(self.text),
        )

        # Refresh the session
        await self.session._refresh()


MultiLineTextInput._unique_id_ = "MultiLineTextInput-builtin"



"""
Pretty-strings a traceback. The result looks very similar to Python's default,
but is colored and just tweaked in general.
"""

import dataclasses
import html
import io
import itertools
import linecache
import os
import sys
import traceback
import typing as t
from pathlib import Path

import path_imports
import revel

import rio

__all__ = [
    "format_exception_revel",
    "format_exception_html",
    "FormatStyle",
    "print_exception",
]


def print_exception(
    error: BaseException, *, relpath: Path | None = None
) -> None:
    text = format_exception_revel(error, relpath=relpath)
    revel.print(text)


@dataclasses.dataclass
class FormatStyle:
    bold: str
    nobold: str
    dim: str
    nodim: str
    yellow: str
    noyellow: str
    red: str
    nored: str
    escape: t.Callable[[str], str]


def _handle_syntax_error(err: SyntaxError) -> traceback.FrameSummary:
    """
    Syntax errors are very special snowflakes and need separate treatment. This
    creates a FrameSummary for a SyntaxError, handling differences between
    Python versions.
    """
    filename = err.filename or "<unknown>"

    # TODO: Is there a better way to do this? Since Python obviously isn't
    # keeping the arguments consistent, this could potentially break every
    # single Python version.
    if sys.version_info < (3, 11):
        return traceback.FrameSummary(
            filename=filename,
            lineno=err.lineno,
            name="<module>",
            line=err.text,
            locals=None,
        )
    else:
        return traceback.FrameSummary(
            filename=filename,
            lineno=err.lineno,
            end_lineno=err.end_lineno,
            colno=None if err.offset is None else err.offset - 1,
            end_colno=None if err.end_offset is None else err.end_offset - 1,
            name="<module>",
            line=err.text,
            locals=None,
        )


def remove_rio_internals_from_traceback(
    tb: list[traceback.FrameSummary],
) -> t.Sequence[traceback.FrameSummary]:
    # Skip frames which are internal to rio (or libraries used by rio) until we
    # hit the first non-rio frame. Then include everything.
    rio_root = rio.__file__
    assert rio_root.endswith(os.sep + "__init__.py")
    rio_root = rio_root.removesuffix(os.sep + "__init__.py")

    def predicate(frame: traceback.FrameSummary) -> bool:
        return (
            frame.filename.startswith((rio_root, "<frozen importlib"))
            or frame.filename == path_imports.__file__
        )

    return list(itertools.dropwhile(predicate, tb))


def _format_single_exception_raw(
    out: t.IO[str],
    err: BaseException,
    *,
    include_header: bool,
    style: FormatStyle,
    relpath: Path | None,
) -> None:
    """
    Format a single exception and write it to the output stream.
    """
    tb_list = traceback.extract_tb(err.__traceback__)

    # Syntax errors need special handling. Convert them to something that
    # behaves more like a regular one.
    if isinstance(err, SyntaxError):
        tb_list.append(_handle_syntax_error(err))

    tb_list = remove_rio_internals_from_traceback(tb_list)

    # TODO: Add special handling for recursion errors. Instead of printing the
    # same frame 1000 times, print a message like "Last 5 frames repeated 200
    # times".

    # Lead-in
    if include_header:
        out.write(
            f"{style.dim}Traceback (most recent call last):{style.nodim}\n"
        )

    # Walk all frames and format them
    for frame in tb_list:
        assert frame.lineno is not None

        # Make paths relative to `relpath` if they're contained within
        frame_path = Path(frame.filename)
        if relpath and frame_path.is_absolute():
            try:
                frame_path = frame_path.relative_to(relpath)
            except ValueError:
                pass

        # File location
        out.write(
            f"  {style.dim}File{style.nodim} {style.yellow}{style.escape(str(frame_path))}{style.noyellow}"
            f"{style.dim}, {style.nodim}line {frame.lineno}{style.noyellow}{style.dim}, in {style.escape(frame.name)}{style.nodim}\n"
        )

        # Display the source code from this line
        #
        # Insanely, the line in the frame has been stripped. Refetch it.
        source_line = linecache.getline(frame.filename, frame.lineno)
        if source_line:
            # If this is the last line, highlight the error
            if (
                frame is tb_list[-1]
                and hasattr(frame, "colno")
                and hasattr(frame, "end_colno")
                and frame.colno is not None  # type: ignore
                and frame.end_colno is not None  # type: ignore
            ):
                start_col = frame.colno  # type: ignore

                if (
                    hasattr(frame, "end_lineno")
                    and frame.end_lineno is not None  # type: ignore
                    and frame.end_lineno > frame.lineno  # type: ignore
                ):
                    end_col = len(source_line) - 1  # -1 to exclude the \n
                else:
                    end_col = frame.end_colno  # type: ignore

                before = style.escape(source_line[:start_col].lstrip())
                error = style.escape(source_line[start_col:end_col])
                after = style.escape(source_line[end_col:].rstrip())
                formatted_line = (
                    f"{before}{style.red}{error}{style.nored}{after}"
                )
            else:
                formatted_line = style.escape(source_line.strip())

            # Write the line
            out.write(f"    {formatted_line}\n")

    # Exception type and message
    out.write("\n")
    out.write(
        f"{style.bold}{style.red}{type(err).__name__}{style.nored}{style.nobold}"
    )

    error_message = str(err)

    if error_message:
        out.write(f"{style.bold}: {style.escape(error_message)}{style.nobold}")


def format_exception_raw(
    err: BaseException,
    *,
    style: FormatStyle,
    relpath: Path | None = None,
) -> str:
    """
    Format an exception into a pretty string with the given style.
    """

    def format_inner(current_err: BaseException) -> None:
        # Chain to the cause or context if there is one
        if current_err.__cause__ is not None:
            format_inner(current_err.__cause__)
            out.write("\n\n")
            out.write(
                "The above exception was the direct cause of the following:\n\n"
            )
            include_header = False
        elif current_err.__context__ is not None:
            format_inner(current_err.__context__)
            out.write("\n\n")
            out.write(
                "During handling of the above exception, another exception occurred:\n\n"
            )
            include_header = False
        else:
            include_header = True

        # Format the exception
        _format_single_exception_raw(
            out,
            current_err,
            include_header=include_header,
            style=style,
            relpath=relpath,
        )

    # Format
    out = io.StringIO()
    format_inner(err)
    return out.getvalue()


def format_exception_revel(
    err: BaseException,
    *,
    relpath: Path | None = None,
) -> str:
    """
    Format an exception using revel's styling.
    """

    # Prepare the style
    style = FormatStyle(
        bold="[bold]",
        nobold="[/]",
        dim="[dim]",
        nodim="[/]",
        yellow="[yellow]",
        noyellow="[/]",
        red="[red]",
        nored="[/]",
        escape=revel.escape,
    )

    # Format the exception
    return format_exception_raw(
        err,
        style=style,
        relpath=relpath,
    )


def format_exception_html(
    err: BaseException,
    *,
    relpath: Path | None = None,
) -> str:
    """
    Format an exception into HTML with appropriate styling.
    """

    # Prepare the style
    style = FormatStyle(
        bold='<span class="rio-traceback-bold">',
        nobold="</span>",
        dim='<span class="rio-traceback-dim">',
        nodim="</span>",
        yellow='<span class="rio-traceback-yellow">',
        noyellow="</span>",
        red='<span class="rio-traceback-red">',
        nored="</span>",
        escape=html.escape,
    )

    # Format the exception
    result = format_exception_raw(
        err,
        style=style,
        relpath=relpath,
    )

    # HTML-ify newlines
    return result.replace("\n", "<br>")



from __future__ import annotations

# <additional-imports>
import rio

from .. import components as comps
from .. import data_models, theme

# </additional-imports>


# <component>
class BlogMinor(rio.Component):
    """
    A component to display a single blog post in a smaller format.


    ## Attributes:

    `blog_post`: Represents the blog post data including title, description,
    image, and other metadata.

    `spacing`: The spacing between blog posts, used to calculate image width for
    desktop layout.
    """

    blog_post: data_models.BlogPost
    spacing: float = 4

    def build(self) -> rio.Component:
        """
        Builds the BlogMinor component.

        The layout adjusts based on the device type:
        - Desktop: Larger images and font sizes, calculated dynamically based on spacing.
        - Mobile: Smaller, static image widths and font sizes.
        """
        # Get the user's device type
        device = self.session[data_models.PageLayout].device
        # Determine device type and adjust sizes accordingly
        if device == "desktop":
            # Dynamically calculate image width for desktop layout
            min_img_width = (80 - 2 * self.spacing) / 3
            blog_post_title_font_size = 1.4
            blog_post_description_text_style = theme.DARK_TEXT_BIGGER

        else:
            min_img_width = (
                self.session.window_width - 2
            )  # substraction of margin_x
            blog_post_title_font_size = 1.2
            blog_post_description_text_style = theme.DARKER_TEXT

        # Content container for the blog post
        content = rio.Column(spacing=2)

        # Add the image
        content.add(
            rio.Rectangle(
                fill=rio.ImageFill(
                    rio.URL(self.blog_post.image),
                    fill_mode="zoom",
                ),
                # Add a fine border around the image
                stroke_width=0.1,
                stroke_color=self.session.theme.neutral_color,
                corner_radius=self.session.theme.corner_radius_medium,
                # Set the cursor to a pointer on hover
                cursor=rio.CursorStyle.POINTER,
                # Set alignment and size of rectangle
                align_x=0,  # Left align
                align_y=0,  # Top align
                min_height=15,  # Fixed height
                min_width=min_img_width,  # Fixed width
            ),
        )

        # Add the content Column, containing the category, title, description,
        # author image and date
        content.add(
            rio.Column(
                # Category button, does not redirect to a page
                comps.OutlinedColoredButton(
                    text=self.blog_post.category,
                    corner_radius=self.session.theme.corner_radius_small,
                    font_size=0.8,
                    font_weight="normal",
                    margin_x_row=0.5,
                    margin_y_row=0.3,
                    align_x=0,
                ),
                # Add the title
                rio.Text(
                    self.blog_post.title,
                    style=rio.TextStyle(
                        font_size=blog_post_title_font_size,
                        font_weight="bold",
                    ),
                ),
                # Add the description
                rio.Text(
                    self.blog_post.description,
                    overflow="wrap",
                    style=blog_post_description_text_style,
                ),
                rio.Row(
                    # Display the author image as a circle
                    rio.Rectangle(
                        # author image
                        fill=rio.ImageFill(
                            self.session.assets / self.blog_post.author_image
                        ),
                        corner_radius=99999,  # Fully rounded for circular appearance
                        cursor=rio.CursorStyle.POINTER,  # Pointer cursor for interactivity
                        stroke_width=0.1,  # Thin border around the circle
                        stroke_color=theme.NEUTRAL_COLOR_BRIGHTER,  # Light border color
                        hover_stroke_color=self.session.theme.primary_color,
                        min_height=2,
                        min_width=2,
                        align_x=0,
                        align_y=0.5,
                    ),
                    # Display the date of the blog post
                    rio.Text(
                        self.blog_post.date,
                        style=theme.DARKER_TEXT,
                        align_y=0.5,  # Vertically center-align the text
                    ),
                    spacing=1,  # Space between image and text
                    align_x=0,  # Align row to the left
                ),
                align_x=0,  # Align column content to the left
                align_y=0.5,  # Vertically center-align the column
                min_width=20,  # Minimum width for the column
                spacing=1,  # Space between elements
            )
        )

        # Return the clickable blog post as a link
        return rio.Link(
            content=rio.Rectangle(
                content=content,
                fill=rio.Color.TRANSPARENT,
                cursor=rio.CursorStyle.POINTER,
            ),
            # Navigate to the blog post's URL when clicked
            target_url=str(self.session.active_page_url)
            # + "/adventure-in-the-swiss-alps",
            + self.blog_post.url_segment,
            # TODO: change fixed string to + self.blog_post.url_segment if all
            # blog posts have a page
        )


# </component>



import abc
import dataclasses

from .component import Component
from .fundamental_component import FundamentalComponent

__all__ = [
    "KeyboardFocusableComponent",
    "KeyboardFocusableFundamentalComponent",
]


class KeyboardFocusableComponent(Component, abc.ABC):
    """
    ## Attributes

    `auto_focus`: Whether this component should receive the keyboard focus when
        it is created.

    ## Metadata

    `public`: False
    """

    _: dataclasses.KW_ONLY
    auto_focus: bool = False

    @abc.abstractmethod
    async def grab_keyboard_focus(self) -> None:
        """
        ## Metadata

        `public`: False
        """
        raise NotImplementedError


class KeyboardFocusableFundamentalComponent(
    KeyboardFocusableComponent, FundamentalComponent
):
    async def grab_keyboard_focus(self) -> None:
        """
        ## Metadata

        `public`: False
        """
        await self.session._remote_set_keyboard_focus(self._id_)



from __future__ import annotations

import asyncio
import dataclasses
import sys
import typing as t
import warnings
import weakref
from collections import defaultdict

import introspection
import typing_extensions as te

import rio

from . import event, global_state, inspection
from .dataclass import RioDataclassMeta, class_local_fields, internal_field
from .state_properties import StateProperty
from .utils import I_KNOW_WHAT_IM_DOING
from .warnings import RioPotentialMistakeWarning

__all__ = ["ComponentMeta"]


C = t.TypeVar("C", bound="rio.Component")


# For some reason vscode doesn't understand that this class is a
# `@te.dataclass_transform`, so we'll annotate it again...
@te.dataclass_transform(
    eq_default=False,
    field_specifiers=(internal_field, dataclasses.field),
)
class ComponentMeta(RioDataclassMeta):
    # Cache for the set of all `StateProperty` instances in this class
    _state_properties_: dict[str, StateProperty]

    # Maps event tags to the methods that handle them. The methods aren't bound
    # to the instance yet, so make sure to pass `self` when calling them
    #
    # The assigned value is needed so that the `Component` class itself has a
    # valid value. All subclasses override this value in `__init_subclass__`.
    _rio_event_handlers_: defaultdict[
        event.EventTag, list[tuple[t.Callable, t.Any]]
    ]

    # Whether this component class is built into Rio, rather than user defined,
    # or from a library.
    _rio_builtin_: bool

    def __init__(cls, *args, **kwargs):
        # Is this class built into Rio?
        cls._rio_builtin_ = cls.__module__.startswith("rio.")

        # Run sanity checks
        if not cls._rio_builtin_:
            try:
                init_func = vars(cls)["__init__"]
            except KeyError:
                pass
            else:
                if init_func not in I_KNOW_WHAT_IM_DOING:
                    warnings.warn(
                        f"`{cls.__name__}` has a custom `__init__` method. This"
                        f" can lead to subtle bugs, so it's usually better to"
                        f" let rio create the constructor automatically. If you"
                        f" need to run code during initialization, use"
                        f" `__post_init__` instead. Otherwise, if you are"
                        f" absolutely sure this is okay, you can silence this"
                        f" warning by adding a `@rio.i_know_what_im_doing`"
                        f" decorator to the `__init__` method.",
                        RioPotentialMistakeWarning,
                    )

        super().__init__(*args, **kwargs)

        # Replace all properties with custom state properties
        cls._initialize_state_properties()

        # Inherit event handlers from parents
        cls._rio_event_handlers_ = defaultdict(list)

        for base in cls.__bases__:
            if not isinstance(base, ComponentMeta):
                continue

            for event_tag, handlers in base._rio_event_handlers_.items():
                cls._rio_event_handlers_[event_tag].extend(handlers)

        # Add events from this class itself
        for member in vars(cls).values():
            if not callable(member):
                continue

            try:
                events = member._rio_events_  # type: ignore
            except AttributeError:
                continue

            for event_tag, args in events.items():
                for arg in args:
                    cls._rio_event_handlers_[event_tag].append((member, arg))

    def _initialize_state_properties(cls) -> None:
        """
        Spawn `StateProperty` instances for all annotated properties in this
        class.
        """
        all_parent_state_properties: dict[str, StateProperty] = {}

        for base in reversed(cls.__bases__):
            if isinstance(base, ComponentMeta):
                all_parent_state_properties.update(base._state_properties_)  # type: ignore (wtf?)

        cls._state_properties_ = all_parent_state_properties

        annotations: dict = vars(cls).get("__annotations__", {})
        module = sys.modules[cls.__module__]

        for field_name, field in class_local_fields(cls).items():
            # Skip internal fields
            if not field.state_property:
                continue

            # Create the StateProperty
            # readonly = introspection.typing.has_annotation(annotation, Readonly
            readonly = False  # TODO

            state_property = StateProperty(
                field_name, readonly, annotations[field_name], module
            )
            setattr(cls, field_name, state_property)

            # Add it to the set of all state properties for rapid lookup
            cls._state_properties_[field_name] = state_property

    @introspection.mark.does_not_alter_signature
    def __call__(
        cls: type[C],  # type: ignore
        *args: object,
        **kwargs: object,
    ) -> C:
        # Inject the session before calling the constructor
        # Fetch the session this component is part of
        if global_state.currently_building_session is None:
            raise RuntimeError(
                "Components can only be created inside of `build` methods."
            )

        # Remap deprecated parameter names to new ones
        args, kwargs = cls._remap_constructor_arguments_(args, kwargs)

        component: C = object.__new__(cls)

        session = global_state.currently_building_session
        component._session_ = session

        # Create a unique ID for this component
        component._id_ = session._next_free_component_id
        session._next_free_component_id += 1

        component._properties_assigned_after_creation_ = set()

        # Call `__init__`
        component.__init__(*args, **kwargs)
        component._init_called_ = True

        # Some components (like `Grid`) manually mark some properties as
        # explicitly set in their `__init__`, so we must use `.update()` instead
        # of an assignment
        component._properties_set_by_creator_.update(
            inspection.get_explicitly_set_state_property_names(
                component,
                args,
                kwargs,
            )
        )

        # If the component has a `key`, register it
        if component.key is not None:
            if component.key in global_state.key_to_component:
                raise RuntimeError(
                    f'Multiple components share the key "{component.key}": {global_state.key_to_component[component.key]} and {component}'
                )

            global_state.key_to_component[component.key] = component

        # Keep track of this component's existence
        #
        # Components must be known by their id, so any messages addressed to
        # them can be passed on correctly.
        session._weak_components_by_id[component._id_] = component

        session._register_dirty_component(
            component,
            include_children_recursively=False,
        )

        # Some events need attention right after the component is created
        for event_tag, event_handlers in component._rio_event_handlers_.items():
            # Don't register an empty list of handlers, since that would
            # still slow down the session
            if not event_handlers:
                continue

            # Page changes are handled by the session. Register the handler
            if event_tag == event.EventTag.ON_PAGE_CHANGE:
                callbacks = tuple(handler for handler, unused in event_handlers)
                session._page_change_callbacks[component] = callbacks

            # Window resizes are handled by the session. Register the handler
            elif event_tag == event.EventTag.ON_WINDOW_SIZE_CHANGE:
                callbacks = tuple(handler for handler, unused in event_handlers)
                session._on_window_size_change_callbacks[component] = callbacks

            # The `periodic` event needs a task to work in
            elif event_tag == event.EventTag.PERIODIC:
                for callback, period in event_handlers:
                    session.create_task(
                        _periodic_event_worker(
                            weakref.ref(component), callback, period
                        ),
                        name=f"`rio.event.periodic` event worker for {component}",
                    )

        # Call `_rio_post_init` for every class in the MRO
        for base in reversed(type(component).__mro__):
            try:
                post_init = vars(base)["_rio_post_init"]
            except KeyError:
                continue

            post_init(component)

        component._properties_assigned_after_creation_.clear()

        return component


async def _periodic_event_worker(
    weak_component: weakref.ReferenceType[rio.Component],
    handler: t.Callable,
    period: float,
) -> None:
    # Get a handle on the session
    try:
        sess = weak_component().session  # type: ignore
    except AttributeError:
        return

    # Keep running for as long as the component exists
    while True:
        # Wait for the next tick
        await asyncio.sleep(period)

        # Wait until there's an active connection to the client. We won't run
        # code periodically if we aren't sure whether the client will come back.
        await sess._is_connected_event.wait()

        # Does the component still exist?
        component = weak_component()

        if component is None:
            return

        # Call the handler
        await component.call_event_handler(handler, component)
        await sess._refresh()

        # Drop the reference to the component
        component = None



from __future__ import annotations

import dataclasses
import types
import typing as t
import weakref

import introspection.typing
import revel

if t.TYPE_CHECKING:
    from .components import Component


__all__ = [
    "StateProperty",
    "AttributeBinding",
    "AttributeBindingMaker",
    "PendingAttributeBinding",
]


class StateProperty:
    """
    StateProperties act like regular properties, with additional considerations:

    - When a state property is assigned to, the component owning it is marked as
      dirty in the session

    - State properties have the ability to share their value with other state
      property instances. If state property `A` is assigned to state property
      `B`, then `B` creates a `StateBinding` and any future access to `B` will
      be routed to `A` instead:

    ```python
    class MyComponent(Component):
        foo_text = "Hello"

        def build(self) -> Component:
            return Bar(bar_text=self.bind().foo_text)  # Note `self.bind()` instead of `self`
    ```
    """

    def __init__(
        self,
        name: str,
        readonly: bool,
        raw_annotation: introspection.types.TypeAnnotation,
        module: types.ModuleType,
    ):
        self.name = name
        self.readonly = readonly

        self._module = module
        self._raw_annotation = raw_annotation
        self._resolved_annotation: introspection.typing.TypeInfo | None = None

    def _annotation_as_string(self) -> str:
        if isinstance(self._raw_annotation, str):
            return self._raw_annotation

        return introspection.typing.annotation_to_string(self._raw_annotation)

    def __get__(
        self,
        instance: Component | None,
        owner: type | None = None,
    ) -> object:
        # If accessed through the class, rather than instance, return the
        # StateProperty itself
        if instance is None:
            return self

        # Otherwise get the value assigned to the property in the component
        # instance
        try:
            value = vars(instance)[self.name]
        except KeyError:
            raise AttributeError(self.name) from None

        # If the value is a binding return the binding's value
        if isinstance(value, AttributeBinding):
            return value.get_value()

        # Otherwise return the value
        return value

    def __set__(self, instance: Component, value: object) -> None:
        if self.readonly:
            cls_name = type(instance).__name__
            raise AttributeError(
                f"Cannot assign to readonly property {cls_name}.{self.name}"
            )

        assert not isinstance(value, StateProperty), (
            f"You're still using the old attribute binding syntax for {instance} {self.name}"
        )

        instance._properties_assigned_after_creation_.add(self.name)

        # Look up the stored value
        instance_vars = vars(instance)
        try:
            local_value = instance_vars[self.name]
        except KeyError:
            # If no value is currently stored, that means this component is
            # currently being instantiated. We may have to create a state
            # binding.
            if isinstance(value, PendingAttributeBinding):
                value = self._create_attribute_binding(instance, value)
        else:
            # If a value is already stored, that means this is a re-assignment.
            # Which further means it's an assignment outside of `__init__`.
            # Which is not a valid place to create a attribute binding.
            if isinstance(value, PendingAttributeBinding):
                raise RuntimeError(
                    "Attribute bindings can only be created when calling the component constructor"
                )

            # Delegate to the binding if it exists
            if isinstance(local_value, AttributeBinding):
                local_value.set_value(value)
                return

        # Otherwise set the value directly and mark the component as dirty
        instance_vars[self.name] = value

        instance._session_._register_dirty_component(
            instance,
            include_children_recursively=False,
        )

    def _create_attribute_binding(
        self,
        component: Component,
        request: PendingAttributeBinding,
    ) -> AttributeBinding:
        # In order to create a `StateBinding`, the owner's attribute
        # must also be a binding
        binding_owner = request._component_
        binding_owner_vars = vars(binding_owner)

        owner_binding = binding_owner_vars[request._state_property_.name]

        if not isinstance(owner_binding, AttributeBinding):
            owner_binding = AttributeBinding(
                owning_component_weak=weakref.ref(binding_owner),
                owning_property=self,
                is_root=True,
                parent=None,
                value=owner_binding,
                children=weakref.WeakSet(),
            )
            binding_owner_vars[request._state_property_.name] = owner_binding

        # Create the child binding
        child_binding = AttributeBinding(
            owning_component_weak=weakref.ref(component),
            owning_property=self,
            is_root=False,
            parent=owner_binding,
            value=None,
            children=weakref.WeakSet(),
        )
        owner_binding.children.add(child_binding)

        return child_binding

    def __repr__(self) -> str:
        return f"<{type(self).__name__} {self.name}>"


@dataclasses.dataclass(eq=False)
class AttributeBinding:
    # Weak reference to the component containing this binding
    owning_component_weak: t.Callable[[], Component | None]

    # The state property whose value this binding is
    owning_property: StateProperty

    # Each binding is either the root-most binding, or a child of another
    # binding. This value is True if this binding is the root.
    is_root: bool

    parent: AttributeBinding | None
    value: object | None

    children: weakref.WeakSet[AttributeBinding] = dataclasses.field(
        default_factory=weakref.WeakSet
    )

    def get_value(self) -> object:
        if self.is_root:
            return self.value

        assert self.parent is not None
        return self.parent.get_value()

    def set_value(self, value: object) -> None:
        # Delegate to the parent, if any
        if self.parent is not None:
            self.parent.set_value(value)
            return

        # Otherwise this is the root-most binding. Set the value
        self.value = value

        # Then recursively mark all children as dirty
        self.recursively_mark_children_as_dirty()

    def recursively_mark_children_as_dirty(self) -> None:
        to_do: list[AttributeBinding] = [self]

        while to_do:
            cur = to_do.pop()
            owning_component = cur.owning_component_weak()

            if owning_component is not None:
                owning_component._session_._register_dirty_component(
                    owning_component,
                    include_children_recursively=False,
                )

            to_do.extend(cur.children)


class AttributeBindingMaker:
    """
    Helper class returned by `Component.bind()`. Used to create attribute
    bindings.
    """

    def __init__(self, component: Component):
        self.component = component

    def __getattribute__(self, name: str) -> object:
        component = super().__getattribute__("__dict__")["component"]
        component_cls = type(component)

        if name not in component_cls._state_properties_:
            raise AttributeError

        state_property = getattr(component_cls, name)
        return PendingAttributeBinding(component, state_property)


class PendingAttributeBinding:
    # This is not a dataclasses because it makes pyright do nonsense
    def __init__(self, component: Component, state_property: StateProperty):
        self._component_ = component
        self._state_property_ = state_property

    def _get_error_message(self, operation: str) -> str:
        return f"You attempted to use `{operation}` on a pending attribute binding. This is not supported. Attribute bindings are an instruction for rio to synchronize the state of two components. They do not have a value. For more information, see https://rio.dev/docs/howto/attribute-bindings"

    def _warn_about_incorrect_usage(self, operation: str) -> None:
        revel.warning(self._get_error_message(operation))

    def __getattr__(self, name: str):
        operation = f".{name}"
        self._warn_about_incorrect_usage(operation)
        raise AttributeError(self._get_error_message(operation))

    def __getitem__(self, item: object):
        operation = f"[{item!r}]"
        self._warn_about_incorrect_usage(operation)
        raise TypeError(self._get_error_message(operation))

    def __add__(self, other):
        self._warn_about_incorrect_usage("+")
        return NotImplemented

    def __sub__(self, other):
        self._warn_about_incorrect_usage("-")
        return NotImplemented

    def __mul__(self, other):
        self._warn_about_incorrect_usage("*")
        return NotImplemented

    def __div__(self, other):
        self._warn_about_incorrect_usage("//")
        return NotImplemented

    def __truediv__(self, other):
        self._warn_about_incorrect_usage("/")
        return NotImplemented

    def __call__(self, other):
        self._warn_about_incorrect_usage("__call__")
        raise TypeError(f"{type(self).__name__} objects are not callable")

    def __repr__(self) -> str:
        self._warn_about_incorrect_usage("__repr__")
        return f"<PendingAttributeBinding for {self._component_!r}.{self._state_property_.name}>"
